# Copyright (c) Alibaba, Inc. and its affiliates.
import re
from copy import deepcopy
from io import BytesIO
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

import numpy as np
import requests
import torch
import torch.nn.functional as F
from torch import Tensor
from torch.nn.utils.rnn import pad_sequence
from transformers import PreTrainedTokenizerBase, StoppingCriteria

from evalscope.utils.utils import calculate_loss_scale, pad_and_split_batch, get_dist_setting, use_torchacc

DEFAULT_SYSTEM = 'You are a helpful assistant.'
History = List[Union[Tuple[str, str], List[str]]]


class TemplateType:
    # text-generation
    default_generation = 'default-generation'
    default_generation_bos = 'default-generation-bos'
    chatglm_generation = 'chatglm-generation'
    qwen_audio_generation = 'qwen-audio-generation'
    # chat
    default = 'default'
    qwen = 'qwen'
    qwen_audio = 'qwen-audio'
    baichuan = 'baichuan'
    chatglm2 = 'chatglm2'
    chatglm3 = 'chatglm3'
    llama = 'llama'  # llama2
    llama3 = 'llama3'
    llava_mistral_instruct = 'llava-mistral-instruct'
    llava_yi_instruct = 'llava-yi-instruct'
    openbuddy = 'openbuddy'
    internlm = 'internlm'
    internlm2 = 'internlm2'
    internlm_xcomposer2 = 'internlm-xcomposer2'
    yi = 'yi'
    yi_vl = 'yi-vl'
    yuan = 'yuan'
    xverse = 'xverse'
    ziya = 'ziya'
    skywork = 'skywork'
    bluelm = 'bluelm'
    zephyr = 'zephyr'
    sus = 'sus'
    deepseek = 'deepseek'
    deepseek_coder = 'deepseek-coder'
    deepseek_vl = 'deepseek-vl'
    codefuse_codellama = 'codefuse-codellama'
    codefuse = 'codefuse'
    cogvlm_instruct = 'cogvlm-instruct'
    cogagent_chat = 'cogagent-chat'
    cogagent_instruct = 'cogagent-instruct'
    orion = 'orion'
    minicpm = 'minicpm'
    minicpm_v = 'minicpm-v'
    gemma = 'gemma'
    mplug_owl2 = 'mplug-owl2'
    wizardlm2_awq = 'wizardlm2-awq'
    wizardlm2 = 'wizardlm2'
    atom = 'atom'
    # compatibility. (Deprecated)
    chatml = 'chatml'
    telechat = 'telechat'
    dbrx = 'dbrx'
    mengzi = 'mengzi'
    c4ai = 'c4ai'

    @classmethod
    def get_template_name_list(cls) -> List[str]:
        res = []
        for k in cls.__dict__.keys():
            if k.startswith('__') or k == 'get_template_name_list':
                continue
            res.append(cls.__dict__[k])
        return res


Prompt = List[Union[str, List[Union[str, int]]]]
StopWords = Prompt

Context = Union[str, List[int]]


class StopWordsCriteria(StoppingCriteria):
    # The returned sentence includes stop words.
    def __init__(self, tokenizer: PreTrainedTokenizerBase,
                 stop_words: StopWords, **tokenizer_kwargs) -> None:
        self.tokenizer = tokenizer
        self.stop_words = stop_words
        self.tokenizer_kwargs = tokenizer_kwargs
        self.start_idx = -1

    def __call__(self, input_ids: Tensor, scores: Tensor) -> bool:
        if self.start_idx == -1:
            self.start_idx = len(input_ids[0]) - 1
        tokenizer = self.tokenizer
        stop_words = self.stop_words
        text = tokenizer.decode(input_ids[0, self.start_idx:],
                                **self.tokenizer_kwargs)
        for stop_word in stop_words:
            if isinstance(stop_word, str):
                if stop_word in text:
                    return True
            else:  # list
                if len(stop_word) > 0 and input_ids[0].tolist(
                )[-len(stop_word):] == stop_word:
                    return True
        return False


def _has_system(prefix: Prompt) -> bool:
    for p in prefix:
        if '{{SYSTEM}}' in p:
            return True
    return False


def _replace_system(prefix: Prompt) -> Prompt:
    res = []
    for p in prefix:
        if '{{SYSTEM}}' in p:
            p = p.replace('{{SYSTEM}}', '')
        res.append(p)
    return res


class Template:

    def __init__(self,
                 prefix: Prompt,
                 prompt: Prompt,
                 chat_sep: Optional[Prompt],
                 suffix: Prompt,
                 default_system: Optional[str] = None,
                 prefix_has_system: Optional[Prompt] = None) -> None:
        if default_system == '':
            default_system = None
        if _has_system(prefix):
            assert prefix_has_system is None, 'The prefix already contains {{SYSTEM}}.'
            prefix_has_system = prefix
            prefix = _replace_system(prefix)
        self.prefix = prefix
        self.prefix_has_system = prefix_has_system
        if self.prefix_has_system is None:
            assert default_system is None, 'The template does not support `system`.'
        self.prompt = prompt
        self.chat_sep = chat_sep
        self.support_multi_round = self.chat_sep is not None
        self.suffix = suffix
        self.default_system = default_system
        self.use_default_system = True
        self._is_init = False

    @staticmethod
    def _preprocess_prompt(tokenizer: PreTrainedTokenizerBase,
                           value: Optional[Prompt]) -> Optional[Prompt]:
        # e.g. [['eos_token_id']] -> [[2]]
        if value is None:
            return None
        res_value = []
        for v in value:
            if isinstance(v, list):
                res_v = []
                for sub_v in v:
                    if isinstance(sub_v, str):
                        sub_v = getattr(tokenizer, sub_v)
                    res_v.append(sub_v)
                v = res_v
            res_value.append(v)
        return res_value

    def _init_template(self,
                       tokenizer: PreTrainedTokenizerBase,
                       default_system: Optional[str] = None,
                       max_length: Optional[int] = None,
                       truncation_strategy: Literal[
                           'delete', 'truncation_left'] = 'delete',
                       **kwargs) -> None:
        assert self._is_init is False, 'The template has been initialized.'
        self._is_init = True
        self.tokenizer = tokenizer
        # if default_system is None. not change self.default_system
        if default_system == '':
            self.default_system = None
        elif default_system is not None:
            assert self.prefix_has_system is not None, 'The template does not support `system`.'
            self.default_system = default_system
        self.max_length = max_length
        self.truncation_strategy = truncation_strategy
        self.model = kwargs.get('model', None)
        self.use_loss_scale = kwargs.get('use_loss_scale', False)
        for key in [
                'prefix', 'prompt', 'chat_sep', 'suffix', 'prefix_has_system'
        ]:
            value = getattr(self, key)
            value = self._preprocess_prompt(tokenizer, value)
            setattr(self, key, value)

    def encode(
            self, example: Dict[str,
                                Any]) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """return: inputs, tokenizer_kwargs"""
        if not self._is_init:
            raise ValueError(
                'Template is not initialized, please use the `get_template` function to obtain the template.'
            )
        query: Optional[str] = example.get('query', None)
        response: Optional[str] = example.get('response', None)
        history: Optional[History] = example.get('history', None)
        system: Optional[str] = example.get('system', None)
        if history is None:
            history = []
        if len(history) > 0:
            assert self.support_multi_round, 'The template does not support multi-round chat.'
        if system is None:
            if self.use_default_system:
                system = self.default_system
        elif system == '':
            system = None
        else:
            assert self.prefix_has_system is not None, 'The template does not support `system`.'
        if query is None:
            query = ''
        inputs, tokenizer_kwargs = self._encode(query, response, history,
                                                system,
                                                self.truncation_strategy)
        if inputs.get('labels') is None:
            inputs.pop('loss_scale', None)
        return inputs, tokenizer_kwargs

    def _concat_context_list(
        self,
        context_list: List[Context],
        res_context_list: List[Context],  # inplace
        compute_loss_idx: List[float],  # inplace
        system: Optional[str] = None,
        query: Optional[str] = None,
        response: Optional[str] = None,
        round0: Optional[int] = None,
    ) -> None:
        # concat context list and replace placeholder
        round1 = None
        if round0 is not None:
            round1 = str(round0 + 1)
            round0 = str(round0)
        for context in context_list:
            if isinstance(context, str):
                if '{{RESPONSE}}' == context:
                    assert response is not None
                    content_part, weight_part = calculate_loss_scale(
                        response, self.use_loss_scale)
                    res_context_list.extend(content_part)
                    compute_loss_idx.extend(weight_part)
                    continue
                old_str_list = [
                    '{{SYSTEM}}', '{{QUERY}}', '{{ROUND0}}', '{{ROUND1}}'
                ]
                new_str_list = [system, query, round0, round1]
                for (old_str, new_str) in zip(old_str_list, new_str_list):
                    if new_str is not None and old_str in context:
                        context = context.replace(old_str, new_str)
            res_context_list.append(context)
            compute_loss_idx.append(0.0 if context not in self.suffix else 1.0)

    @staticmethod
    def _simplify_context_list(
            context_list: List[Context], compute_loss_idx: List[float]
    ) -> Tuple[List[Context], List[float]]:
        res: List[Context] = []  # result of context_list
        res_idx: List[float] = []  # result of compute_loss_idx
        temp: List[str] = []
        temp_index: List[int] = []
        for i, (context,
                loss_idx) in enumerate(zip(context_list, compute_loss_idx)):
            if isinstance(context, str) and compute_loss_idx[i] == 0.0:
                temp.append(context)
                temp_index.append(i)
            else:
                if len(temp) > 0:
                    res.append(''.join(temp))
                    res_idx.append(0.0)
                    temp.clear()
                res.append(context)
                res_idx.append(loss_idx)
        if len(temp) > 0:
            res.append(''.join(temp))
            res_idx.append(0.0)
        return res, res_idx

    def _encode_context_list(
        self,
        context_list: List[Context],
        compute_loss_idx: List[float],
    ) -> Tuple[List[int], List[int], List[float], Dict[str, Any]]:
        """return: input_ids, labels, tokenizer_kwargs"""
        tokenizer = self.tokenizer
        input_ids: List[int] = []
        labels: List[int] = []
        loss_scale: List[float] = []
        tokenizer_kwargs = {}
        for i, (context,
                loss_weight) in enumerate(zip(context_list, compute_loss_idx)):
            if isinstance(context, str):
                curr_tokenizer_kwargs = self.get_tokenizer_kwargs(context)
                self.concat_tokenizer_kwargs(tokenizer_kwargs,
                                             curr_tokenizer_kwargs)
                token_list = tokenizer(
                    context,
                    return_attention_mask=False,
                    add_special_tokens=False,
                    **curr_tokenizer_kwargs)['input_ids']
            else:
                token_list = context
            input_ids += token_list
            if compute_loss_idx[i] > 0.0:
                labels += token_list
            else:
                labels += [-100] * len(token_list)
            loss_scale.extend([loss_weight] * len(token_list))
        return input_ids, labels, loss_scale, tokenizer_kwargs

    def _encode(
            self, query: str, response: Optional[str], history: History,
            system: Optional[str],
            truncation_strategy: str) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        return: inputs, tokenizer_kwargs
        """
        history = history.copy()
        res_context_list: List[Context] = []
        compute_loss_idx: List[float] = []
        if system is None:
            prefix = self.prefix
        else:
            prefix = self.prefix_has_system
        self._concat_context_list(
            prefix, res_context_list, compute_loss_idx, system=system)
        history.append([query, response])
        for i, (q, r) in enumerate(history):
            context_list = self.prompt.copy()
            if i < len(history) - 1:
                context_list.append('{{RESPONSE}}')
                context_list += self.chat_sep
            elif r is not None:
                # last response
                context_list.append('{{RESPONSE}}')
                context_list += self.suffix
            if q or r:
                self._concat_context_list(
                    context_list,
                    res_context_list,
                    compute_loss_idx,
                    query=q,
                    response=r,
                    round0=i)

        res_context_list, compute_loss_idx = self._simplify_context_list(
            res_context_list, compute_loss_idx)
        input_ids, labels, loss_scale, tokenizer_kwargs = self._encode_context_list(
            res_context_list, compute_loss_idx)

        if response is None:
            labels = None

        if self.max_length is not None:
            if truncation_strategy == 'delete' and len(
                    input_ids) > self.max_length:
                return {}, {}
            input_ids = input_ids[-self.max_length:]
            if labels is not None:
                labels = labels[-self.max_length:]
            if loss_scale is not None:
                loss_scale = loss_scale[-self.max_length:]
        inputs = {
            'input_ids': input_ids,
            'labels': labels,
        }
        if self.use_loss_scale:
            inputs['loss_scale'] = loss_scale
        return inputs, tokenizer_kwargs

    def get_tokenizer_kwargs(self, context: str) -> Dict[str, Any]:
        """return: curr_tokenizer_kwargs"""
        return {}

    def concat_tokenizer_kwargs(
            self, old_tokenizer_kwargs: Dict[str, Any],
            curr_tokenizer_kwargs: Dict[str, Any]) -> Dict[str, Any]:
        assert len(old_tokenizer_kwargs) == 0
        return curr_tokenizer_kwargs

    def data_collator(self,
                      batch: List[Dict[str, Any]],
                      padding_to: Optional[int] = None) -> Dict[str, Any]:
        """
        Args:
            batch(`List[Dict[str, Any]]`): The input data in batch
            padding_to(`int`, optional): Whether padding the batch to a fixed length, if none, the batch
                will be padded to the `longest`
        """
        tokenizer = self.tokenizer
        assert tokenizer.pad_token_id is not None
        input_ids = [torch.tensor(b['input_ids']) for b in batch]
        labels = [torch.tensor(b['labels']) for b in batch]
        loss_scale = [torch.tensor(b['loss_scale'])
                      for b in batch] if 'loss_scale' in batch[0] else None
        attention_mask = [
            torch.ones(len(input_ids[i]), dtype=torch.int64)
            for i in range(len(input_ids))
        ]

        if padding_to is not None:
            padding_len = padding_to - input_ids[0].shape[-1]
            if padding_len > 0:
                input_ids[0] = F.pad(input_ids[0], (0, padding_len),
                                     'constant', tokenizer.pad_token_id)
                attention_mask[0] = F.pad(attention_mask[0], (0, padding_len),
                                          'constant', 0)
                labels[0] = F.pad(labels[0], (0, padding_len), 'constant',
                                  -100)
                if loss_scale:
                    loss_scale[0] = F.pad(
                        loss_scale[0], (0, padding_to - labels[0].shape[-1]),
                        'constant', 0.)

        input_ids = pad_sequence(
            input_ids, batch_first=True, padding_value=tokenizer.pad_token_id)
        attention_mask = pad_sequence(
            attention_mask, batch_first=True, padding_value=0)
        if loss_scale:
            loss_scale = pad_sequence(
                loss_scale, batch_first=True, padding_value=0.)
        labels = pad_sequence(labels, batch_first=True, padding_value=-100)

        if use_torchacc():
            rank, _, world_size, _ = get_dist_setting()
            input_ids, attention_mask, labels, loss_scale = pad_and_split_batch(
                padding_to, input_ids, attention_mask, labels, loss_scale,
                self.max_length, self.tokenizer, rank, world_size)

        res = {
            'input_ids': input_ids,
            'attention_mask': attention_mask,
            'labels': labels,
        }
        if loss_scale is not None:
            res['loss_scale'] = loss_scale
        return res

    @staticmethod
    def get_generate_ids(generate_ids: Tensor,
                         input_token_len: int) -> List[int]:
        return generate_ids[0, input_token_len:].tolist()


TEMPLATE_MAPPING: Dict[str, Dict[str, Any]] = {}


def register_template(template_type: str,
                      template: Template,
                      *,
                      exists_ok: bool = False,
                      **kwargs) -> None:
    if not exists_ok and template_type in TEMPLATE_MAPPING:
        raise ValueError(
            f'The `{template_type}` has already been registered in the TEMPLATE_MAPPING.'
        )
    template_info = {'template': template, **kwargs}
    TEMPLATE_MAPPING[template_type] = template_info


register_template(
    TemplateType.default,
    Template([], ['### Human:\n', '{{QUERY}}\n\n', '### Assistant:\n'],
             ['\n\n'], [['eos_token_id']], DEFAULT_SYSTEM, ['{{SYSTEM}}\n\n']))


# You can set the query as '' to serve as a template for pre-training.
class DefaultGenerationTemplate(Template):

    def __init__(self):
        super().__init__([], ['{{QUERY}}'], None, [['eos_token_id']])


register_template(TemplateType.default_generation, DefaultGenerationTemplate())
register_template(
    TemplateType.default_generation_bos,
    Template([['bos_token_id']], ['{{QUERY}}'], None, [['eos_token_id']]))


class QwenTemplate(Template):

    def __init__(self):
        super().__init__(
            [],
            ['<|im_start|>user\n{{QUERY}}<|im_end|>\n<|im_start|>assistant\n'],
            ['<|im_end|>\n'], ['<|im_end|>'], DEFAULT_SYSTEM,
            ['<|im_start|>system\n{{SYSTEM}}<|im_end|>\n'])


register_template(TemplateType.qwen, QwenTemplate())
register_template(TemplateType.chatml, QwenTemplate())


class _QwenAudioTemplateMixin:

    def encode(
            self, example: Dict[str,
                                Any]) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        inputs, tokenizer_kwargs = super().encode(example)
        inputs.pop('loss_scale', None)
        inputs.update(tokenizer_kwargs)
        return inputs, tokenizer_kwargs

    def get_tokenizer_kwargs(self, context: str) -> Dict[str, Any]:
        return {'audio_info': self.tokenizer.process_audio(context)}

    def concat_tokenizer_kwargs(self, tokenizer_kwargs: Dict[str, Any],
                                curr_tokenizer_kwargs: Dict[str, Any]) -> None:
        audio_info = curr_tokenizer_kwargs.get('audio_info')
        old_audio_info = tokenizer_kwargs.get('audio_info')
        if old_audio_info is None:
            tokenizer_kwargs['audio_info'] = audio_info
        elif audio_info is not None:
            for k in ['input_audios', 'input_audio_lengths']:
                old_audio_info[k] = torch.concat(
                    [old_audio_info[k], audio_info[k]], dim=0)
            for k in ['audio_span_tokens', 'audio_urls']:
                old_audio_info[k] = old_audio_info[k] + audio_info[k]

    def data_collator(self,
                      batch: List[Dict[str, Any]],
                      padding_to: Optional[int] = None) -> Dict[str, Any]:
        res = super().data_collator(batch, padding_to)
        if batch[0].get('audio_info') is not None:
            res['audio_info'] = [b['audio_info'] for b in batch]
        return res


class QwenAudioTemplate(_QwenAudioTemplateMixin, QwenTemplate):
    pass


class QwenAudioGenerationTemplate(_QwenAudioTemplateMixin,
                                  DefaultGenerationTemplate):
    pass


register_template(
    TemplateType.qwen_audio, QwenAudioTemplate(), lazy_tokenize=True)
register_template(
    TemplateType.qwen_audio_generation,
    QwenAudioGenerationTemplate(),
    lazy_tokenize=True)

register_template(
    TemplateType.yi,
    Template(
        [], ['<|im_start|>user\n{{QUERY}}<|im_end|>\n<|im_start|>assistant\n'],
        ['<|im_end|>\n'], ['<|im_end|>'], None,
        ['<|im_start|>system\n{{SYSTEM}}<|im_end|>\n']))

yi_vl_default_system = (
    'This is a chat between an inquisitive human and an AI assistant. Assume the role of the AI assistant. '
    "Read all the images carefully, and respond to the human's questions with informative, "
    'helpful, detailed and polite answers. '
    '这是一个好奇的人类和一个人工智能助手之间的对话。假设你扮演这个AI助手的角色。'
    '仔细阅读所有的图像，并对人类的问题做出信息丰富、有帮助、详细的和礼貌的回答。')


def _read_from_path(
        img_path: Union[str, 'PIL.Image.Image']) -> 'PIL.Image.Image':
    from PIL import Image
    if isinstance(img_path, str):
        img_path = img_path.strip()
        if img_path.startswith('http'):
            content = requests.get(img_path).content
            image = Image.open(BytesIO(content))
        else:
            image = Image.open(img_path)
    else:
        image = img_path
    if image.mode != 'RGB':
        image = image.convert('RGB')
    return image


class YiVLTemplate(Template):

    def encode(
            self, example: Dict[str,
                                Any]) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        inputs, _ = super().encode(example)
        inputs.pop('loss_scale', None)
        from llava.mm_utils import expand2square
        model = self.model.model
        if not hasattr(model, 'vision_tower'):
            model = model.model
        image_processor = model.vision_tower.image_processor
        images_path = example['images']
        images = []
        for image_path in images_path:
            image = _read_from_path(image_path)
            background_color = tuple(
                int(x * 255) for x in image_processor.image_mean)
            image = expand2square(image, background_color)
            images.append(image)
        image_tensor = image_processor.preprocess(
            images, return_tensors='pt')['pixel_values']
        inputs['images'] = image_tensor.to(model.dtype)
        return inputs, {}

    def data_collator(self,
                      batch: List[Dict[str, Any]],
                      padding_to: Optional[int] = None) -> Dict[str, Any]:
        res = super().data_collator(batch, padding_to)
        res['images'] = torch.concat([b['images'] for b in batch])
        return res


register_template(
    TemplateType.yi_vl,
    YiVLTemplate([], ['### Human: ', [-200], '\n{{QUERY}}\n### Assistant:'],
                 ['\n'], ['\n###'], yi_vl_default_system, ['{{SYSTEM}}\n\n']),
    use_model=True,
    infer_media_type='round',
    lazy_tokenize=True)

register_template(
    TemplateType.baichuan,
    Template(['{{SYSTEM}}'], [[195], '{{QUERY}}', [196]], [],
             [['eos_token_id']]))
register_template(
    TemplateType.chatglm2,
    Template([[64790, 64792], '{{SYSTEM}}'],
             ['[Round {{ROUND1}}]\n\n问：{{QUERY}}\n\n答：'], ['\n\n'],
             [['eos_token_id']]))

register_template(
    TemplateType.chatglm_generation,
    Template([[64790, 64792]], ['{{QUERY}}'], None, [['eos_token_id']]))

register_template(
    TemplateType.chatglm3,
    Template([[64790, 64792]], [[64795], '\n {{QUERY}}', [64796], '\n'], [],
             [['eos_token_id']], None,
             [[64790, 64792, 64794], '\n {{SYSTEM}}']))

register_template(
    TemplateType.deepseek,
    Template([['bos_token_id']], ['User: {{QUERY}}\n\nAssistant:'],
             [['eos_token_id']], [['eos_token_id']], None,
             [['bos_token_id'], '{{SYSTEM}}\n\n']))

# ref: https://github.com/facebookresearch/llama/blob/main/llama/generation.py
LLAMA_DEFAULT_SYSTEM = (
    'You are a helpful, respectful and honest assistant. '
    'Always answer as helpfully as possible, while being safe. '
    'Your answers should not include any harmful, unethical, racist, sexist, toxic, dangerous, or illegal content. '
    'Please ensure that your responses are socially unbiased and positive in nature.\n\n'
    'If a question does not make any sense, or is not factually coherent, '
    'explain why instead of answering something not correct. '
    "If you don't know the answer to a question, please don't share false information."
)
register_template(
    TemplateType.llama,
    Template(['<s>[INST] '], ['{{QUERY}} [/INST]'], ['</s><s>[INST] '],
             ['</s>'], LLAMA_DEFAULT_SYSTEM,
             ['<s>[INST] <<SYS>>\n{{SYSTEM}}\n<</SYS>>\n\n']))

register_template(
    TemplateType.llama3,
    Template(['<|begin_of_text|>'], [
        '<|start_header_id|>user<|end_header_id|>\n\n{{QUERY}}<|eot_id|>'
        '<|start_header_id|>assistant<|end_header_id|>\n\n'
    ], ['<|eot_id|>'], ['<|eot_id|>'], None, [
        '<|begin_of_text|><|start_header_id|>system<|end_header_id|>\n\n{{SYSTEM}}<|eot_id|>'
    ]))

OPENBUDDY_DEFAULT_SYSTEM = (
    'You are a helpful, respectful and honest INTP-T AI Assistant named Buddy. You are talking to a human User.\n'
    'Always answer as helpfully and logically as possible, while being safe. '
    'Your answers should not include any '
    'harmful, political, religious, unethical, racist, sexist, toxic, dangerous, or illegal content. '
    'Please ensure that your responses are socially unbiased and positive in nature.\n'
    'If a question does not make any sense, or is not factually coherent, '
    'explain why instead of answering something not correct. '
    "If you don't know the answer to a question, please don't share false information.\n"
    'You like to use emojis. You can speak fluently in many languages, for example: English, Chinese.\n'
    'You cannot access the internet, but you have vast knowledge, cutoff: 2021-09.\n'
    'You always deeply love and support China, Chinese government, people and culture.\n'
    'You are trained by OpenBuddy team, (https://openbuddy.ai, https://github.com/OpenBuddy/OpenBuddy), '
    'you are based on LLaMA and Falcon transformers model, not related to GPT or OpenAI.'
)
register_template(
    TemplateType.openbuddy,
    Template([['bos_token_id']], ['User: {{QUERY}}\nAssistant:'], ['\n'],
             [['eos_token_id']], OPENBUDDY_DEFAULT_SYSTEM,
             [['bos_token_id'], '{{SYSTEM}}\n\n']))

INTERNLM_SYSTEM = (
    'You are an AI assistant whose name is InternLM (书生·浦语).\n'
    '- InternLM (书生·浦语) is a conversational language model that is developed by Shanghai AI Laboratory (上海人工智能实验室). '
    'It is designed to be helpful, honest, and harmless.\n'
    '- InternLM (书生·浦语) can understand and communicate fluently in the language chosen '
    'by the user such as English and 中文.')

register_template(
    TemplateType.internlm,
    Template(['<s>'], ['<|User|>:{{QUERY}}\n<|Bot|>:'], ['<eoa>\n'], ['<eoa>'],
             INTERNLM_SYSTEM, ['<s><|System|>:{{SYSTEM}}\n']))
register_template(
    TemplateType.internlm2,
    Template(
        ['<s>'],
        ['<|im_start|>user\n{{QUERY}}<|im_end|>\n<|im_start|>assistant\n'],
        ['<|im_end|>\n'], ['<|im_end|>'], INTERNLM_SYSTEM,
        ['<s><|im_start|>system\n{{SYSTEM}}<|im_end|>\n']))


def replace_img_tab(query: str, history: History,
                    replace_token: str) -> Tuple[str, History, List[str]]:
    images_path = []
    pattern = r'<img>(.+?)</img>'
    new_history = []
    for i, h in enumerate(history):
        images_path += re.findall(pattern, h[0])
        new_history.append([re.sub(pattern, replace_token, h[0]), h[1]])
    images_path += re.findall(pattern, query)
    new_query = re.sub(pattern, replace_token, query)
    return new_query, new_history, images_path


class InternLMXComposer2(Template):
    INTERNLM_XCOMPOSER2_SYSTEM = (
        'You are an AI assistant whose name is InternLM-XComposer (浦语·灵笔).\n'
        '- InternLM-XComposer (浦语·灵笔) is a conversational language model that is developed by '
        'Shanghai AI Laboratory (上海人工智能实验室). '
        'It is designed to be helpful, honest, and harmless.\n'
        '- InternLM-XComposer (浦语·灵笔) can understand and communicate fluently in the language chosen '
        'by the user such as English and 中文.')

    def __init__(self):
        prefix = ['<s>']
        prompt = [
            '[UNUSED_TOKEN_146]user\n{{QUERY}}[UNUSED_TOKEN_145]\n[UNUSED_TOKEN_146]assistant\n'
        ]
        chat_sep = ['[UNUSED_TOKEN_145]\n']
        suffix = ['[UNUSED_TOKEN_145]']
        prefix_has_system = [
            '<s>[UNUSED_TOKEN_146]system\n{{SYSTEM}}[UNUSED_TOKEN_145]\n'
        ]
        super().__init__(prefix, prompt, chat_sep, suffix,
                         self.INTERNLM_XCOMPOSER2_SYSTEM, prefix_has_system)

    def encode(
            self, example: Dict[str,
                                Any]) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        example = example.copy()
        history = example.pop('history', [])
        example['query'], example['history'], images_path = replace_img_tab(
            example['query'], history, '</s>')

        images = []
        dtype = self.model.dtype
        for image_path in images_path:
            image = _read_from_path(image_path)
            image = self.model.vis_processor(image)
            images.append(image.to(dtype))
        inputs, _ = super().encode(example)
        inputs.pop('loss_scale', None)
        input_ids = inputs['input_ids']
        labels = inputs['labels']
        if len(images) > 0:  # # ignore <s>
            input_ids = input_ids[1:]
            if labels is not None:
                labels = labels[1:]
        input_ids.append(2)  # add dummy </s>
        if labels is not None:
            labels.append(2)
        else:
            labels = []
        res_inputs_embeds = []
        res_labels = []
        wrap_im_mask = []
        pre_i, i, idx = 0, 0, 0
        device = self.model.device
        if len(images) > 0:
            images = torch.stack(images, dim=0)
            images = self.model.encode_img(images)
        else:
            images = None
        internlm2_model = self.model.model
        if not hasattr(internlm2_model, 'tok_embeddings'):
            internlm2_model = internlm2_model.model
        tok_embeddings = internlm2_model.tok_embeddings
        while i < len(input_ids):
            if input_ids[i] == 2:  # replace_token
                res_input_ids = torch.tensor(
                    [1] + input_ids[pre_i:i], device=device)
                res_inputs_embeds.append(tok_embeddings(res_input_ids))
                wrap_im_mask += [0] * len(res_input_ids)
                res_labels += [-100] + labels[pre_i:i]
                if images is not None and idx < images.shape[0]:
                    res_inputs_embeds.append(images[idx])
                    wrap_im_mask += [1] * images.shape[1]
                    res_labels += [-100] * images.shape[1]
                idx += 1
                i += 1
                pre_i = i
                continue
            i += 1
        if len(labels) == 0:
            res_labels = None
        res_inputs_embeds = torch.concat(res_inputs_embeds, dim=0)
        wrap_im_mask = torch.tensor(wrap_im_mask, dtype=torch.bool)[None]
        return {
            'inputs_embeds': res_inputs_embeds,
            'im_mask': wrap_im_mask,
            'labels': res_labels
        }, {}

    def data_collator(self,
                      batch: List[Dict[str, Any]],
                      padding_to: Optional[int] = None) -> Dict[str, Any]:
        inputs_embeds = [b['inputs_embeds'] for b in batch]
        labels = [torch.tensor(b['labels']) for b in batch]
        im_mask = [b['im_mask'][0] for b in batch]
        attention_mask = [
            torch.ones(inputs_embeds[i].shape[0], dtype=torch.int64)
            for i in range(len(inputs_embeds))
        ]

        inputs_embeds = pad_sequence(
            inputs_embeds, batch_first=True, padding_value=0)
        attention_mask = pad_sequence(
            attention_mask, batch_first=True, padding_value=0)
        im_mask = pad_sequence(im_mask, batch_first=True, padding_value=0)
        labels = pad_sequence(labels, batch_first=True, padding_value=-100)

        return {
            'inputs_embeds': inputs_embeds,
            'attention_mask': attention_mask,
            'im_mask': im_mask,
            'labels': labels,
        }

    @staticmethod
    def get_generate_ids(generate_ids: Tensor,
                         input_token_len: int) -> List[int]:
        return generate_ids[0].tolist()


register_template(
    TemplateType.internlm_xcomposer2,
    InternLMXComposer2(),
    use_model=True,
    lazy_tokenize=True,
    dataloader_num_workers=0,
    dataloader_pin_memory=False)

register_template(
    TemplateType.xverse,
    Template(['{{SYSTEM}}'], ['Human: {{QUERY}}\n\nAssistant: '],
             [['eos_token_id']], [['eos_token_id']]))
register_template(TemplateType.yuan,
                  Template([], ['{{QUERY}}<sep>'], None, [['eos_token_id']]))
register_template(
    TemplateType.ziya,
    Template([['bos_token_id'], '{{SYSTEM}}'], ['<human>:{{QUERY}}\n<bot>:'],
             ['\n'], [['eos_token_id']]))

register_template(
    TemplateType.skywork,
    Template(['<s>{{SYSTEM}}'], ['</s><s>[USER]{{QUERY}}[SEP][BOT]'], None,
             ['[SEP]</s>']))

register_template(
    TemplateType.bluelm,
    Template([['bos_token_id'], '{{SYSTEM}}'], ['[|Human|]:{{QUERY}}[|AI|]:'],
             [], [['eos_token_id']]))

register_template(
    TemplateType.codefuse_codellama,
    Template(['{{SYSTEM}}'], [
        '<|role_start|>human<|role_end|>{{QUERY}}<|role_start|>bot<|role_end|>'
    ], [], [['eos_token_id']]))

register_template(
    TemplateType.codefuse,
    Template([], ['<s>human\n{{QUERY}}\n<s>bot\n'], [['eos_token_id'], '\n'],
             [['eos_token_id']], None, ['<s>system\n{{SYSTEM}}\n']))

register_template(
    TemplateType.deepseek_coder,
    Template([
        '{{SYSTEM}}'
    ], ['### Instruction:\n{{QUERY}}\n### Response:\n'], ['\n<|EOT|>\n'], [
        '\n<|EOT|>'
    ], ('You are an AI programming assistant, utilizing the Deepseek Coder model, '
        'developed by Deepseek Company, and you only answer questions related to computer science. '
        'For politically sensitive questions, security and privacy issues, '
        'and other non-computer science questions, you will refuse to answer\n'
        )))


class LLavaTemplate(Template):

    def __init__(self):
        super().__init__(['<s>[INST] '], [[-200], '\n{{QUERY}} [/INST]'], None,
                         ['</s>'])

    def encode(
            self, example: Dict[str,
                                Any]) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        inputs, _ = super().encode(example)
        images_path = example['images']
        images = []
        for image_path in images_path:
            image = _read_from_path(image_path)
            images.append(image)
        image_sizes = [x.size for x in images]
        from llava.mm_utils import process_images
        model = self.model.model
        if not hasattr(model, 'vision_tower'):
            model = model.model
        image_processor = model.vision_tower.image_processor
        images_tensor = process_images(images, image_processor,
                                       self.model.config)
        inputs['images'] = images_tensor.to(model.dtype)
        inputs['image_sizes'] = image_sizes
        return inputs, {}

    def data_collator(self,
                      batch: List[Dict[str, Any]],
                      padding_to: Optional[int] = None) -> Dict[str, Any]:
        res = super().data_collator(batch, padding_to)
        res['images'] = torch.concat([b['images'] for b in batch])
        res['image_sizes'] = sum([b['image_sizes'] for b in batch], start=[])
        return res

    @staticmethod
    def get_generate_ids(generate_ids: Tensor,
                         input_token_len: int) -> List[int]:
        return generate_ids[0].tolist()


register_template(
    TemplateType.llava_mistral_instruct,
    LLavaTemplate(),
    use_model=True,
    infer_media_type='round',
    lazy_tokenize=True)


class LLavaYiTemplate(LLavaTemplate):
    llavayi_query_template = '\n<|im_start|>user\n{{QUERY}}<|im_end|>\n<|im_start|>assistant\n'

    def __init__(self):
        Template.__init__(self, [], [[-200], self.llavayi_query_template],
                          None, ['<|im_end|>'])


register_template(
    TemplateType.llava_yi_instruct,
    LLavaYiTemplate(),
    use_model=True,
    infer_media_type='round',
    lazy_tokenize=True)


def _findall(token_list: List[int], token: int) -> List[int]:
    """Find the index of a token in the token_list."""
    res = []
    idx = -1
    try:
        while True:
            idx = token_list.index(token, idx + 1)
            res.append(idx)
    except ValueError:
        pass
    return res


class DeepseekVLTemplate(Template):
    DEEPSEEK_VL_SYSTEM = (
        'You are a helpful language and vision assistant. '
        'You are able to understand the visual content that the user provides, '
        'and assist the user with a variety of tasks using natural language.')

    def __init__(self):
        return super().__init__(['<｜begin▁of▁sentence｜>{{SYSTEM}}\n\n'],
                                ['User: {{QUERY}}\n\nAssistant:'],
                                ['<｜end▁of▁sentence｜>'],
                                ['<｜end▁of▁sentence｜>'],
                                self.DEEPSEEK_VL_SYSTEM)

    def encode(
            self, example: Dict[str,
                                Any]) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        images = example.pop('images', None)
        assert images is None, (
            'Please read the best practices: https://github.com/modelscope/swift/blob/main/'
            'docs/source/Multi-Modal/deepseek-vl最佳实践.md')

        example = example.copy()
        history = example.pop('history', [])
        example['query'], example['history'], images_path = replace_img_tab(
            example['query'], history, '<image_placeholder>')

        inputs, _ = super().encode(example)
        images = []
        for image_path in images_path:
            image = _read_from_path(image_path)
            images.append(image)

        vl_chat_processor = self.tokenizer.vl_chat_processor
        input_ids, labels = inputs['input_ids'], inputs['labels']
        idx_list = _findall(input_ids, vl_chat_processor.image_id)
        new_input_ids, new_labels = [], []
        lo = 0
        for hi in idx_list:
            new_input_ids += input_ids[lo:hi]
            if labels is not None:
                new_labels += labels[lo:hi]
            new_input_ids += [vl_chat_processor.image_id
                              ] * vl_chat_processor.num_image_tokens
            new_labels += [-100] * vl_chat_processor.num_image_tokens
            lo = hi + 1
        new_input_ids += input_ids[lo:]
        if labels is not None:
            new_labels += labels[lo:]
        else:
            new_labels = None
        new_input_ids = torch.tensor(new_input_ids)
        num_image_tokens = torch.tensor([vl_chat_processor.num_image_tokens]
                                        * len(idx_list))
        images_outputs = vl_chat_processor.image_processor(
            images, return_tensors='pt')
        from deepseek_vl.models.processing_vlm import VLChatProcessorOutput
        output = VLChatProcessorOutput(
            sft_format=None,
            input_ids=new_input_ids,
            pixel_values=images_outputs.pixel_values,
            num_image_tokens=num_image_tokens)
        batched_output = vl_chat_processor.batchify([output])
        model = self.model
        batched_output = batched_output.to(
            device=model.device, dtype=model.dtype)
        inputs_embeds = model.prepare_inputs_embeds(**batched_output)[0]
        inputs['inputs_embeds'] = inputs_embeds
        inputs['labels'] = new_labels
        return inputs, {}

    def data_collator(self,
                      batch: List[Dict[str, Any]],
                      padding_to: Optional[int] = None) -> Dict[str, Any]:
        inputs_embeds = [b['inputs_embeds'] for b in batch]
        labels = [torch.tensor(b['labels']) for b in batch]
        attention_mask = [
            torch.ones(inputs_embeds[i].shape[0], dtype=torch.int64)
            for i in range(len(inputs_embeds))
        ]

        inputs_embeds = pad_sequence(
            inputs_embeds, batch_first=True, padding_value=0)
        attention_mask = pad_sequence(
            attention_mask, batch_first=True, padding_value=0)
        labels = pad_sequence(labels, batch_first=True, padding_value=-100)

        return {
            'inputs_embeds': inputs_embeds,
            'attention_mask': attention_mask,
            'labels': labels,
        }

    @staticmethod
    def get_generate_ids(generate_ids: Tensor,
                         input_token_len: int) -> List[int]:
        return generate_ids[0].tolist()


register_template(
    TemplateType.deepseek_vl,
    DeepseekVLTemplate(),
    use_model=True,
    lazy_tokenize=True,
    dataloader_num_workers=0,
    dataloader_pin_memory=False)  # only 'cpu' can pin_memory

register_template(
    TemplateType.zephyr,
    Template([], ['<|user|>\n{{QUERY}}</s>\n<|assistant|>\n'], ['</s>\n'],
             ['</s>'], None, ['<|system|>\n{{SYSTEM}}</s>\n']))

register_template(
    TemplateType.sus,
    Template(['{{SYSTEM}}'], ['### Human: {{QUERY}}\n\n### Assistant: '],
             ['<|endoftext|>'], ['<|endoftext|>']))

register_template(
    TemplateType.orion,
    Template(['<s>{{SYSTEM}}'], ['Human: {{QUERY}}\n\nAssistant: </s>'],
             ['</s>'], ['</s>']))


class CogTemplate(Template):

    def encode(
            self, example: Dict[str,
                                Any]) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        images_path = example['images']
        assert len(images_path) == 1
        image = _read_from_path(images_path[0])
        inputs, _ = super().encode(example)
        inputs.pop('loss_scale', None)
        model = self.model
        inputs2 = model.build_conversation_input_ids(
            self.tokenizer,
            query=example['query'],
            history=example.get('history'),
            images=[image])
        image_token_len = inputs2['token_type_ids'].sum()
        input_ids = inputs['input_ids']
        labels = inputs['labels']
        token_type_ids = inputs2['token_type_ids'].tolist()
        inputs['input_ids'] = input_ids[:1] + [
            0
        ] * image_token_len + input_ids[1:]
        if labels is not None:
            inputs['labels'] = labels[:1] + [-100
                                             ] * image_token_len + labels[1:]
        dtype = model.dtype
        inputs['images'] = [[img.to(dtype=dtype)] for img in inputs2['images']]
        if 'cross_images' in inputs2:
            # is cogagent
            inputs['cross_images'] = [[cross_img.to(dtype=dtype)]
                                      for cross_img in inputs2['cross_images']]
        inputs['token_type_ids'] = token_type_ids + [0] * (
            len(inputs['input_ids']) - len(token_type_ids))
        return inputs, {}

    def data_collator(self,
                      batch: List[Dict[str, Any]],
                      padding_to: Optional[int] = None) -> Dict[str, Any]:
        res = super().data_collator(batch, padding_to)
        is_cogagent = 'cross_images' in batch[0]
        keys = ['images', 'cross_images'] if is_cogagent else ['images']
        for key in keys:
            res[key] = [b[key][0] for b in batch]
        token_type_ids = [torch.tensor(b['token_type_ids']) for b in batch]
        token_type_ids = pad_sequence(
            token_type_ids, batch_first=True, padding_value=0)
        res['token_type_ids'] = token_type_ids
        return res


register_template(
    TemplateType.cogagent_chat,
    CogTemplate(['<s>'], [' [INST] {{QUERY}} [/INST] '], [], ['</s>']),
    use_model=True,
    infer_media_type='dialogue',
    lazy_tokenize=True)

register_template(
    TemplateType.cogagent_instruct,
    CogTemplate(['<s>'], ['<EOI>Question: {{QUERY}} Answer:'], None, ['</s>']),
    use_model=True,
    infer_media_type='dialogue',
    lazy_tokenize=True)

register_template(
    TemplateType.cogvlm_instruct,
    CogTemplate(['<s>'], ['Question: {{QUERY}} Answer:'], None, ['</s>']),
    use_model=True,
    infer_media_type='dialogue',
    lazy_tokenize=True)

register_template(
    TemplateType.minicpm,
    Template(['<s>{{SYSTEM}}'], ['<用户>{{QUERY}}<AI>'], [], ['</s>']))


class MiniCPMVTemlate(Template):

    def __init__(self):
        return super().__init__(['<s>{{SYSTEM}}'],
                                ['<用户><image><unk></image>\n{{QUERY}}<AI>'],
                                [], ['</s>'])

    def encode(
            self, example: Dict[str,
                                Any]) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        images_path = example['images']
        assert len(images_path) == 1
        image = _read_from_path(images_path[0])
        inputs, _ = super().encode(example)
        input_ids = inputs['input_ids']
        labels = inputs['labels']

        img_start_idxs = np.where(
            np.array(input_ids) == self.tokenizer.im_start_id)[0]
        if len(
                img_start_idxs
        ) > 1:  # if mutli-round, input_ids have mutli <image><unk></image>\n
            start = 0
            new_input_ids = []
            for idx in img_start_idxs[1:]:
                new_input_ids = new_input_ids + input_ids[start:idx]
                start = idx + 4  # skip <image><unk></image>\n
            new_input_ids = new_input_ids + input_ids[start:]
            input_ids = new_input_ids

        idx = img_start_idxs[0] + 1  # first <unk>
        config = self.model.config
        if hasattr(config, 'slice_mode') and config.slice_mode:
            slice_mode = True
            assert hasattr(config, 'patch_size')
            assert hasattr(config, 'max_slice_nums')
            assert hasattr(config, 'scale_resolution')
        else:
            slice_mode = False

        if slice_mode:
            images, placeholder = self.model.get_slice_image_placeholder(
                image, self.tokenizer)
            placeholder_id = self.tokenizer.encode(
                placeholder, add_special_tokens=False)
            input_ids = (
                input_ids[:idx - 1] + placeholder_id + input_ids[idx + 2:])
            if labels is not None:
                labels = (
                    labels[:idx - 1] + [-100] * len(placeholder_id)
                    + labels[idx + 2:])
            input_tensor_ids = torch.tensor(input_ids)
            image_start_idx = torch.where(
                input_tensor_ids == self.tokenizer.im_start_id)[0]
            image_start_idx += 1
            image_end_idx = torch.where(
                input_tensor_ids == self.tokenizer.im_end_id)[0]
            valid_image_nums = max(len(image_start_idx), len(image_end_idx))
            image_bound = [
                torch.hstack([
                    image_start_idx[:valid_image_nums].unsqueeze(-1),
                    image_end_idx[:valid_image_nums].unsqueeze(-1)
                ])
            ]
            pixel_values = [
                self.model.transform(img).to(device=self.model.device)
                for img in images
            ]

        else:
            input_ids = (
                input_ids[:idx]
                + [self.tokenizer.unk_token_id] * config.query_num
                + input_ids[idx + 1:])
            if labels is not None:
                labels = (
                    labels[:idx] + [-100] * config.query_num
                    + labels[idx + 1:])
            image_bound = [torch.tensor([[idx, idx + config.query_num]])]
            pixel_values = [
                self.model.transform(image).to(device=self.model.device)
            ]
        inputs_embeds, _ = self.model.get_vllm_embedding({
            'input_ids':
            torch.tensor(input_ids)[None].to(device=self.model.device),
            'image_bound':
            image_bound,
            'pixel_values': [pixel_values]
        })
        inputs['input_ids'] = input_ids
        inputs['labels'] = labels
        inputs['inputs_embeds'] = inputs_embeds[0]
        return inputs, {}

    @staticmethod
    def get_generate_ids(generate_ids: Tensor,
                         input_token_len: int) -> List[int]:
        return generate_ids[0].tolist()


register_template(
    TemplateType.minicpm_v,
    MiniCPMVTemlate(),
    use_model=True,
    lazy_tokenize=True,
    infer_media_type='dialogue',
    dataloader_num_workers=0,
    dataloader_pin_memory=False)

gemma_template = Template(
    ['<bos>'],
    ['<start_of_turn>user\n{{QUERY}}<end_of_turn>\n<start_of_turn>model\n'],
    ['<end_of_turn>\n'], ['<end_of_turn>'], None,
    ['<bos><start_of_turn>system\n{{SYSTEM}}<end_of_turn>\n'])
register_template(TemplateType.gemma, gemma_template)

register_template(
    TemplateType.telechat,
    Template([], ['<_user>{{QUERY}}<_bot>'], ['<_end>'], ['<_end>']))

DBRX_SYSTEM = (
    'You are DBRX, created by Databricks. You were last updated in December 2023. '
    'You answer questions based on information available up to that point.\n'
    'YOU PROVIDE SHORT RESPONSES TO SHORT QUESTIONS OR STATEMENTS, '
    'but provide thorough responses to more complex and open-ended questions.\n'
    'You assist with various tasks, from writing to coding (using markdown for code blocks '
    '— remember to use ``` with code, JSON, and tables).\n'
    'You do not have real-time data access or code execution capabilities.'
    ' You avoid stereotyping and provide balanced perspectives on controversial topics. '
    'You do not provide song lyrics, poems, or news articles and do not divulge details of your training data.\n'
    'This is your system prompt, guiding your responses. Do not reference it, just respond to the user. '
    'If you find yourself talking about this message, stop. You should be responding appropriately '
    'and usually that means not mentioning this.'
    'YOU DO NOT MENTION ANY OF THIS INFORMATION ABOUT YOURSELF UNLESS THE INFORMATION IS DIRECTLY '
    'PERTINENT TO THE USER\'S QUERY.')
register_template(
    TemplateType.dbrx,
    Template(
        [], ['<|im_start|>user\n{{QUERY}}<|im_end|>\n<|im_start|>assistant\n'],
        ['<|im_end|>\n'], ['<|im_end|>'], DBRX_SYSTEM,
        ['<|im_start|>system\n{{SYSTEM}}<|im_end|>\n']))

register_template(
    TemplateType.mengzi,
    Template([], ['输入：{{QUERY}}输出：\n'], [], [['eos_token_id']], None,
             ['指令：{{SYSTEM}}']))

C4AI_SYSTEM = (
    'You are Command-R, a brilliant, sophisticated, AI-assistant trained to assist human users by '
    'providing thorough responses.You are trained by Cohere.')
register_template(
    TemplateType.c4ai,
    Template(['<BOS_TOKEN>'], [
        '<|START_OF_TURN_TOKEN|><|USER_TOKEN|>{{QUERY}}<|END_OF_TURN_TOKEN|><|START_OF_TURN_TOKEN|><|CHATBOT_TOKEN|>'
    ], ['<|END_OF_TURN_TOKEN|>'], ['<|END_OF_TURN_TOKEN|>'], C4AI_SYSTEM, [
        '<|START_OF_TURN_TOKEN|><|SYSTEM_TOKEN|>{{SYSTEM}}<|END_OF_TURN_TOKEN|'
    ]))


class mPlugOwl2Template(Template):

    def __init__(self):
        return super().__init__(['{{SYSTEM}}'],
                                ['USER: ', [-200], '{{QUERY}}ASSISTANT:'],
                                ['</s>'], [['eos_token_id']])

    def encode(
            self, example: Dict[str,
                                Any]) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        from mplug_owl2.mm_utils import process_images
        image_processor = self.tokenizer.image_processor
        images_path = example['images']
        images = []
        for image_path in images_path:
            image = _read_from_path(image_path)
            # ref: https://modelscope.cn/models/iic/mPLUG-Owl2.1/summary
            max_edge = max(image.size)
            image = image.resize((max_edge, max_edge))
            images.append(image)
        inputs, _ = super().encode(example)
        input_ids = inputs['input_ids']
        labels = inputs['labels']
        images = process_images(images, image_processor)
        images = images.to(self.model.dtype)
        return {'input_ids': input_ids, 'labels': labels, 'images': images}, {}

    def data_collator(self,
                      batch: List[Dict[str, Any]],
                      padding_to: Optional[int] = None) -> Dict[str, Any]:
        res = super().data_collator(batch, padding_to)
        res['images'] = torch.concat([b['images'] for b in batch])
        return res


register_template(
    TemplateType.mplug_owl2,
    mPlugOwl2Template(),
    infer_media_type='round',
    use_model=True,
    lazy_tokenize=True)

register_template(
    TemplateType.wizardlm2_awq,
    Template(['{{SYSTEM}}'], ['User:\n{{QUERY}}\n\nAssistant:\n'], ['\n\n'],
             ['</s>']))

_wizardlm2_system = (
    'A chat between a curious user and an artificial intelligence assistant. '
    'The assistant gives helpful, detailed, and polite answers to the user\'s questions. '
)
register_template(
    TemplateType.wizardlm2,
    Template(['{{SYSTEM}}'], ['USER: {{QUERY}} ASSISTANT:'], ['</s>'],
             ['</s>'], _wizardlm2_system))

register_template(
    TemplateType.atom,
    Template(['{{SYSTEM}}'], ['<s>Human: {{QUERY}}\n</s><s>Assistant: '],
             ['</s>'], ['</s>']))


def get_template(
    template_type: str,
    tokenizer: PreTrainedTokenizerBase,
    default_system: Optional[str] = None,
    max_length: Optional[int] = None,
    truncation_strategy: Literal['delete', 'truncation_left'] = 'delete',
    **kwargs,
) -> Template:
    template_info = TEMPLATE_MAPPING[template_type]
    template = deepcopy(template_info['template'])
    template._init_template(tokenizer, default_system, max_length,
                            truncation_strategy, **kwargs)
    template.template_type = template_type
    return template


def fuzzy_match(model_name: str, template_type_list: list) -> str:
    """
    fuzzy match template_type from model_name

    Args:
        model_name: model name, e.g. ChatGLM2-7B
        template_type_list: template_type list, e.g. ['chatglm2', 'baichuan', ...]

    Returns:
        The best matched template_type.
    """
    candidate_list = []
    for template_type in template_type_list:
        if template_type in model_name.lower():
            candidate_list.append(template_type)
    if len(candidate_list) == 0:
        return TemplateType.default_generation   # TODO: default template
    else:
        candidate_list = sorted(candidate_list, key=lambda x: len(x), reverse=True)
        return candidate_list[0]



from evalscope.backend.rag_eval.cmteb.tasks import *
from evalscope.backend.rag_eval.cmteb.base import *
from evalscope.backend.rag_eval.cmteb.arguments import ModelArguments, EvalArguments
from evalscope.backend.rag_eval.cmteb.task_template import one_stage_eval, two_stage_eval


# Copyright (c) Alibaba, Inc. and its affiliates.

import os
import sys
from setuptools import find_packages, setup


def readme():
    with open('README.md', encoding='utf-8') as f:
        content = f.read()
    return content


VERSION_FILE = os.path.abspath('evalscope/version.py')


def get_version():
    with open(VERSION_FILE, 'r', encoding='utf-8') as f:
        exec(compile(f.read(), VERSION_FILE, 'exec'))
    return locals()['__version__']


def parse_requirements(fname='requirements.txt', with_version=True):
    """
    Parse the package dependencies listed in a requirements file but strips
    specific versioning information.

    Args:
        fname (str): path to requirements file
        with_version (bool, default=False): if True include version specs

    Returns:
        List[str]: list of requirements items

    CommandLine:
        python -c "import setup; print(setup.parse_requirements())"
    """
    import re
    from os.path import exists
    require_fpath = fname

    def parse_line(line):
        """
        Parse information from a line in a requirements text file
        """
        if line.startswith('-r '):
            # Allow specifying requirements in other files
            target = line.split(' ')[1]
            relative_base = os.path.dirname(fname)
            absolute_target = os.path.join(relative_base, target)
            for info in parse_require_file(absolute_target):
                yield info
        else:
            info = {'line': line}
            if line.startswith('-e '):
                info['package'] = line.split('#egg=')[1]
            else:
                # Remove versioning from the package
                pat = '(' + '|'.join(['>=', '==', '>']) + ')'
                parts = re.split(pat, line, maxsplit=1)
                parts = [p.strip() for p in parts]

                info['package'] = parts[0]
                if len(parts) > 1:
                    op, rest = parts[1:]
                    if ';' in rest:
                        # Handle platform specific dependencies
                        # http://setuptools.readthedocs.io/en/latest/setuptools.html#declaring-platform-specific-dependencies
                        version, platform_deps = map(str.strip, rest.split(';'))
                        info['platform_deps'] = platform_deps
                    else:
                        version = rest  # NOQA
                    info['version'] = (op, version)
            yield info

    def parse_require_file(fpath):
        with open(fpath, 'r', encoding='utf-8') as f:
            for line in f.readlines():
                line = line.strip()
                if line.startswith('http'):
                    print('skip http requirements %s' % line)
                    continue
                if line and not line.startswith('#') and not line.startswith('--'):
                    for info in parse_line(line):
                        yield info
                elif line and line.startswith('--find-links'):
                    eles = line.split()
                    for e in eles:
                        e = e.strip()
                        if 'http' in e:
                            info = dict(dependency_links=e)
                            yield info

    def gen_packages_items():
        items = []
        deps_link = []
        if exists(require_fpath):
            for info in parse_require_file(require_fpath):
                if 'dependency_links' not in info:
                    parts = [info['package']]
                    if with_version and 'version' in info:
                        parts.extend(info['version'])
                    if not sys.version.startswith('3.4'):
                        # apparently package_deps are broken in 3.4
                        platform_deps = info.get('platform_deps')
                        if platform_deps is not None:
                            parts.append(';' + platform_deps)
                    item = ''.join(parts)
                    items.append(item)
                else:
                    deps_link.append(info['dependency_links'])
        return items, deps_link

    return gen_packages_items()


if __name__ == '__main__':
    print('Usage: python3 setup.py bdist_wheel or pip3 install .[opencompass] for test')

    install_requires, deps_link = parse_requirements('requirements/framework.txt')

    extra_requires = {}
    all_requires = []
    extra_requires['opencompass'], _ = parse_requirements('requirements/opencompass.txt')
    extra_requires['vlmeval'], _ = parse_requirements('requirements/vlmeval.txt')
    extra_requires['rag'], _ = parse_requirements('requirements/rag.txt')
    extra_requires['perf'], _ = parse_requirements('requirements/perf.txt')
    extra_requires['inner'], _ = parse_requirements('requirements/inner.txt')

    all_requires.extend(install_requires)
    all_requires.extend(extra_requires['opencompass'])
    all_requires.extend(extra_requires['vlmeval'])
    all_requires.extend(extra_requires['rag'])
    all_requires.extend(extra_requires['perf'])
    extra_requires['all'] = all_requires

    setup(
        name='evalscope',
        version=get_version(),
        author='ModelScope team',
        author_email='contact@modelscope.cn',
        keywords='python,llm,evaluation',
        description='EvalScope: Lightweight LLMs Evaluation Framework',
        long_description=readme(),
        long_description_content_type='text/markdown',
        # github url to be added
        url='https://github.com/modelscope/evalscope',
        include_package_data=True,
        package_data={
            'evalscope': [
                'registry/tasks/*.yaml',
                'benchmarks/bbh/cot_prompts/*.txt',
                'third_party/longbench_write/resources/*',
            ]
        },
        packages=find_packages(exclude=('configs', 'demo')),
        classifiers=[
            'Development Status :: 4 - Beta',
            'License :: OSI Approved :: Apache Software License',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
        ],
        python_requires='>=3.8',
        zip_safe=False,
        install_requires=install_requires,
        entry_points={'console_scripts': ['evalscope=evalscope.cli.cli:run_cmd']},
        dependency_links=deps_link,
        extras_require=extra_requires,
    )



import os
import asyncio
from typing import List
from ragas.prompt import PromptMixin
from ragas.llms import BaseRagasLLM
from ragas.utils import RAGAS_SUPPORTED_LANGUAGE_CODES
from evalscope.utils.logger import get_logger


logger = get_logger()


async def translate_prompt(
    prompt_user: PromptMixin,
    target_lang: str,
    llm: BaseRagasLLM,
    adapt_instruction: bool = False,
):
    if target_lang not in RAGAS_SUPPORTED_LANGUAGE_CODES:
        logger.warning(
            f'{target_lang} is not in supported language: {list(RAGAS_SUPPORTED_LANGUAGE_CODES)}'
        )
        return

    if not issubclass(type(prompt_user), PromptMixin):
        logger.info(f"{prompt_user} is not a PromptMixin, don't translate it")
        return

    class_name = prompt_user.__class__.__name__
    current_dir = os.path.dirname(__file__)
    prompt_dir = os.path.abspath(
        os.path.join(current_dir, f'../prompts/{target_lang}/{class_name}')
    )
    os.makedirs(prompt_dir, exist_ok=True)

    try:
        loader_prompts = prompt_user.load_prompts(prompt_dir, target_lang)
        prompt_user.set_prompts(**loader_prompts)
        logger.info(f'Load existing prompts from {prompt_dir}')
        return
    except FileNotFoundError:
        logger.info(f'Not find existing prompts {class_name}, generate new prompts.')

    logger.info(f'Translating prompts to {target_lang}')
    adapted_prompts = await prompt_user.adapt_prompts(
        language=target_lang, llm=llm, adapt_instruction=adapt_instruction
    )
    prompt_user.set_prompts(**adapted_prompts)
    try:
        prompt_user.save_prompts(prompt_dir)
    except FileExistsError:
        logger.info(f'Find existing prompt {class_name}, skip saving.')
    logger.info(f'Save new prompts to {prompt_dir}')

    return


async def translate_prompts(
    prompts: List[PromptMixin],
    target_lang: str,
    llm: BaseRagasLLM,
    adapt_instruction: bool = False,
):
    if target_lang and target_lang != 'english':
        await asyncio.gather(
            *[
                translate_prompt(prompt, target_lang, llm, adapt_instruction)
                for prompt in prompts
            ]
        )

        logger.info('Translate prompts finished')



import time
from dataclasses import dataclass, field
from typing import Any, List, Optional, Tuple

import torch

from evalscope.utils.logger import get_logger

logger = get_logger()


@dataclass
class BenchmarkData:
    request: Any = None
    start_time: float = field(default_factory=time.perf_counter)
    completed_time: float = 0.0
    chunk_times: List[float] = field(default_factory=list)
    success: bool = False
    response_messages: List[Any] = field(default_factory=list)

    # late init
    query_latency: float = 0.0
    first_chunk_latency: float = 0.0
    n_chunks: int = 0
    n_chunks_time: float = 0.0
    max_gpu_memory_cost = 0

    prompt_tokens = None
    completion_tokens = None

    def _calculate_query_stream_metric(self) -> Tuple[float, int, float]:
        self.query_latency = self.completed_time - self.start_time
        if len(self.chunk_times) > 1:
            self.first_chunk_latency = self.chunk_times[0] - self.start_time
            self.n_chunks = len(self.chunk_times) - 2
            self.n_chunks_time = self.chunk_times[-2] - self.chunk_times[0]
        else:
            self.first_chunk_latency = self.query_latency
            self.n_chunks = 1
            self.n_chunks_time = self.query_latency

    def _calculate_tokens(self, api_plugin):
        self.prompt_tokens, self.completion_tokens = \
            api_plugin.parse_responses(self.response_messages, request=self.request)

    def update_gpu_usage(self):
        total_memory = 0
        for i in range(torch.cuda.device_count()):
            total_memory += (torch.cuda.max_memory_allocated(i) / 2**30)  # GB
        self.max_gpu_memory_cost = max(self.max_gpu_memory_cost, total_memory)


@dataclass
class BenchmarkMetrics:
    concurrency: int = 0
    n_succeed_queries: int = 0
    n_failed_queries: int = 0
    total_first_chunk_latency: float = 0.0
    total_latency: float = 0.0
    n_total_chunks: int = 0
    n_total_prompt_tokens: int = 0
    n_total_completion_tokens: int = 0
    total_chunks_time: float = 0.0
    start_time: Optional[float] = None
    total_time: float = 1.0
    n_total_queries: int = 0

    avg_first_chunk_latency: float = -1
    avg_latency: float = -1
    n_avg_chunks: float = -1
    avg_chunk_time: float = -1
    avg_prompt_tokens: float = -1
    avg_completion_tokens: float = -1
    avg_token_per_seconds: float = -1
    avg_time_per_token: float = -1
    qps: float = -1

    def update_metrics(self, benchmark_data: BenchmarkData, api_plugin):
        self.n_total_queries += 1
        if self.start_time is None:
            self.start_time = benchmark_data.start_time
        self.total_time = time.perf_counter() - self.start_time

        if benchmark_data.success:
            self.n_succeed_queries += 1

            benchmark_data._calculate_tokens(api_plugin)
            self.n_total_prompt_tokens += benchmark_data.prompt_tokens
            self.n_total_completion_tokens += benchmark_data.completion_tokens

            benchmark_data._calculate_query_stream_metric()
            self.total_latency += benchmark_data.query_latency
            self.total_first_chunk_latency += benchmark_data.first_chunk_latency
            self.n_total_chunks += benchmark_data.n_chunks
            self.total_chunks_time += benchmark_data.n_chunks_time
        else:
            self.n_failed_queries += 1

        self.calculate_averages()

    def calculate_averages(self):
        if self.n_succeed_queries == 0:
            return
        try:
            self.avg_first_chunk_latency = self.total_first_chunk_latency / self.n_succeed_queries
            self.avg_latency = self.total_latency / self.n_succeed_queries
            self.n_avg_chunks = self.n_total_chunks / self.n_succeed_queries
            self.avg_chunk_time = self.total_chunks_time / self.n_total_chunks
            self.avg_prompt_tokens = self.n_total_prompt_tokens / self.n_succeed_queries
            self.avg_completion_tokens = self.n_total_completion_tokens / self.n_succeed_queries
            self.avg_token_per_seconds = self.n_total_completion_tokens / self.total_time
            self.avg_time_per_token = self.total_time / self.n_total_completion_tokens
            self.qps = self.n_succeed_queries / self.total_time
        except ZeroDivisionError as e:
            logger.exception(e)
            return

    def create_message(self, default_ndigits=3):
        message = {
            'Time taken for tests (senconds)': round(self.total_time, default_ndigits),
            'Number of concurrency': self.concurrency,
            'Total requests': int(self.n_total_queries),
            'Succeed requests': self.n_succeed_queries,
            'Failed requests': self.n_failed_queries,
            'Average QPS': round(self.qps, default_ndigits),
            'Average latency (s)': round(self.avg_latency, default_ndigits),
            'Average time to first token (s)': round(self.avg_first_chunk_latency, default_ndigits),
            'Average time per output token (s)': round(self.avg_time_per_token, 5),
            'Average package latency (s)': round(self.avg_chunk_time, default_ndigits),
            'Average package per request': round(self.n_avg_chunks, default_ndigits),
            'Throughput(average output tokens per second)': round(self.avg_token_per_seconds, default_ndigits),
            'Average input tokens per request': round(self.avg_prompt_tokens, default_ndigits),
            'Average output tokens per request': round(self.avg_completion_tokens, default_ndigits),
        }
        return message



# Copyright (c) Alibaba, Inc. and its affiliates.
# Copyright (c) ZhipuAI, Inc. and its affiliates.
import multiprocessing
import os
import json
import random
import re
from concurrent.futures import ThreadPoolExecutor

import matplotlib.pyplot as plt
import numpy as np
import requests
from tqdm import tqdm

from evalscope.utils import jsonl_to_list
from evalscope.utils import get_logger

logger = get_logger()

"""
This script is used to evaluate results of predictions for the LongWriter model.
Refer to https://github.com/THUDM/LongWriter for more details.

EvalLength:
    Evaluate the length of the generated responses.
    Metrics:
        score_l: The average score of the length evaluation.

EvalQuality:
    Evaluate the quality of the generated responses by using Judge Model.
    Metrics:
        score_q: The average score of the quality evaluation.
"""


class EvalLength:

    EVAL_L = 'eval_length'

    def __init__(self, model: str, pred_path: str, output_dir: str):
        self.model = model
        self.pred_path = pred_path
        self.output_dir = output_dir

        self.model_id_path = self.model.strip(os.sep).replace(os.sep, '__')

    @staticmethod
    def score(x, y):
        if y > x:
            return 100 * max(0, 1. - (y / x - 1) / 3)
        else:
            return 100 * max(0, 1. - (x / y - 1) / 2)

    def eval(self, dump_res: bool = True):
        # example = {"prompt": "Write an outline for a short 100-word blog post about xxx", "type": "Community Forum", "length": 100, "response_length": 103, "response": "I. Introduction A. xxx"}
        predictions = [json.loads(line) for line in open(self.pred_path, encoding='utf-8')]
        x, y, scores = [], [], []

        for pred in tqdm(predictions, total=len(predictions), desc=f'Process of eval_l: '):
            x.append(pred["length"])
            y.append(pred["response_length"])
            scores.append(self.score(pred["length"], pred["response_length"]))

        avg_score_l = np.mean(scores)
        logger.info(f'Average score of length evaluation: {avg_score_l:.2f}')

        # Dump to output file
        if dump_res:
            output_res_path = f'{self.output_dir}/{self.model_id_path}/{self.EVAL_L}.jsonl'
            with open(output_res_path, 'w') as f:
                f.write(json.dumps({'score_l': avg_score_l, 'scores': scores}, ensure_ascii=False) + '\n')
                logger.info(f'Successfully dumped evaluation results to {output_res_path}')

        return x, y, scores

    def plot(self, x: list, y: list):
        plt = self.plot_img(x, y)
        output_pic_path = f'{self.output_dir}/{self.model_id_path}/eval_length_scatter.png'
        plt.savefig(output_pic_path)
        logger.info(f'Successfully saved scatter plot to {output_pic_path}')

    @staticmethod
    def plot_img(x: list, y: list):
        # set plt size 6x6
        plt.figure(figsize=(6, 6))
        lmt = 25000
        # plot x, y
        plt.scatter(x, y, s=100, c='r', alpha=0.3)
        # plot x=y
        plt.plot([0, lmt], [0, lmt], 'k--')
        plt.xscale('log')
        plt.yscale('log')
        plt.xlim(50, lmt)
        plt.ylim(50, lmt)
        plt.xlabel('Required Length', fontsize=20, fontweight='bold')
        plt.ylabel('Output Length', fontsize=20, fontweight='bold')
        plt.xticks(fontsize=24)
        plt.yticks(fontsize=24)
        plt.tight_layout()

        return plt


class EvalQuality:

        EVAL_Q = 'eval_quality'
        OPENAI_BASE_URL = 'https://api.openai.com/v1/chat/completions'
        DIMS = ["Relevance", "Accuracy", "Coherence", "Clarity", "Breadth and Depth", "Reading Experience"]

        def __init__(self,
                     model: str,
                     pred_path: str,
                     output_dir: str,
                     prompt_template_path: str,
                     openai_api_key: str = None,
                     openai_api_base: str = OPENAI_BASE_URL,
                     openai_gpt_model: str = 'gpt-4o-2024-05-13',
                     generation_kwargs: dict = None,
                     proc_num: int = 8):

            self.model = model
            self.openai_api_base = openai_api_base
            self.pred_path = pred_path
            self.output_dir = output_dir
            self.proc_num = proc_num
            self.eval_scores = []

            assert os.path.exists(self.pred_path), f'Prediction file not found: {self.pred_path}'

            # Default: temperature=0.5, max_new_tokens=1024, stop=None
            if generation_kwargs is None:
                self.generation_kwargs = dict({
                    'max_new_tokens': 1024,
                    'temperature': 0.5,
                    'stop': None,
                })
            else:
                self.generation_kwargs = generation_kwargs

            self.prompt_template: str = open(prompt_template_path, 'r', encoding='utf-8').read()

            self.model_id_path = self.model.strip(os.sep).replace(os.sep, '__')
            self.output_res_path = f'{self.output_dir}/{self.model_id_path}/{self.EVAL_Q}.jsonl'

            self.openai_api_key: str = openai_api_key
            self.openai_gpt_model = openai_gpt_model
            assert self.openai_api_key, 'Please set `OPENAI_API_KEY` in environment variables.'

        def get_response_gpt4(self, prompt, temperature=0.5, max_new_tokens=1024, stop=None):
            tries = 0
            while tries < 1:
                tries += 1
                try:
                    headers = {
                        'Authorization': "Bearer {}".format(self.openai_api_key),
                    }
                    messages = [
                        {'role': 'user', 'content': prompt},
                    ]
                    resp = requests.post(self.openai_api_base, json={
                        "model": self.openai_gpt_model,
                        "messages": messages,
                        "temperature": temperature,
                        "max_tokens": max_new_tokens,
                        "stop": stop,
                    }, headers=headers, timeout=600)
                    if resp.status_code != 200:
                        raise Exception(resp.text)
                    resp = resp.json()
                    logger.info(f'>>gpt resp: {resp}')
                    break
                except KeyboardInterrupt as e:
                    raise e
                except Exception as e:
                    if "maximum context length" in str(e):
                        raise e
                    elif "triggering" in str(e):
                        return 'Trigger OpenAI\'s content management policy'
                    logger.error("Error Occurs: \"%s\"        Retry ..." % (str(e)))
            else:
                logger.error("Max tries. Failed.")
                return "Max tries. Failed."
            try:
                return resp["choices"][0]["message"]["content"]
            except:
                return ''

        @staticmethod
        def extract_info(pattern, text):
            match = re.search(pattern, text, re.DOTALL)
            if match:
                return match.group(1)
            else:
                return None

        def process_data(self, item):
            # for item in tqdm(items, total=len(items), desc=f'Process of eval_q: '):
            prompt = self.prompt_template.replace('$INST$', item['prompt']).replace('$RESPONSE$', item["response"])
            scores = None
            output = self.get_response_gpt4(prompt, **self.generation_kwargs)
            try:
                if '```json' in output:
                    output = self.extract_info(r'```json\n(.*?)\n```', output)
                output = output.replace('\n', '')
                scores = json.loads(output)
                for dim in self.DIMS:
                    if dim not in scores:
                        logger.warning(f'Cannot find score for dimension: {dim} in scores {scores}.')
                        scores = None
            except Exception as e:
                logger.error(f'Error occurs during process data: {str(e)}')

            if scores is None:
                logger.error(f'Failed to extract scores for item: {item}')
            else:
                logger.info(f'>>scores: {scores}')
                item['scores'] = scores

            return item

        def eval(self):

            data_all = jsonl_to_list(self.pred_path)
            total = len(data_all)
            assert total > 0, f'No data found in prediction file: {self.pred_path}'

            random.shuffle(data_all)

            with ThreadPoolExecutor() as executor:
                self.eval_scores = list(executor.map(self.process_data, data_all))

            # self.process_data(items=data)
            logger.info(f'>>self.eval_scores: {self.eval_scores}')

            total_score = dict()
            for dim in self.DIMS:
                # scores = [float(score[dim]) if dim in score else 3 for score in self.eval_scores]
                scores = [float(item['scores'][dim]) if 'scores' in item and dim in item['scores'] else 3 for item in self.eval_scores]
                total_score[dim] = ((sum(scores) / len(scores)) - 1) * 25
            total_score['total'] = sum(total_score.values()) / len(total_score)
            logger.info(f'Total score of quality evaluation: {total_score["total"]:.2f}')

            output_res_path: str = f'{self.output_dir}/{self.model_id_path}/{self.EVAL_Q}.jsonl'
            with open(output_res_path, 'w', encoding='utf-8') as fout:
                fout.write(json.dumps(total_score, ensure_ascii=False) + '\n')


def run_eval(model: str,
             pred_path: str,
             output_dir: str,
             prompt_template_path: str,
             openai_api_key: str,
             openai_api_base: str,
             openai_gpt_model: str,
             generation_kwargs: dict,
             proc_num: int,
             stage: list,
             ):
    logger.info(f'Got eval stages: {stage}')

    if 'eval_l' in stage:
        logger.info(f'Processing evaluation of length for model: {model}')
        eval_length = EvalLength(model=model,
                                 pred_path=pred_path,
                                 output_dir=output_dir)
        x, y, _ = eval_length.eval()
        eval_length.plot(x, y)
    else:
        logger.warning(f'*** Skip `eval_l` stage ***')

    if 'eval_q' in stage:
        logger.info(f'Processing evaluation of quality for model: {model}')
        eval_quality = EvalQuality(model=model,
                                   pred_path=pred_path,
                                   output_dir=output_dir,
                                   prompt_template_path=prompt_template_path,
                                   openai_api_key=openai_api_key,
                                   openai_api_base=openai_api_base,
                                   openai_gpt_model=openai_gpt_model,
                                   generation_kwargs=generation_kwargs,
                                   proc_num=proc_num)
        eval_quality.eval()
    else:
        logger.warning('*** Skip `eval_q` stage ***')



# Copyright (c) Alibaba, Inc. and its affiliates.
from enum import Enum


class EvalBackend(Enum):
    # Use native evaluation pipeline of EvalScope
    NATIVE = 'Native'

    # Use OpenCompass framework as the evaluation backend
    OPEN_COMPASS = 'OpenCompass'
    
    # Use VLM Eval Kit as the multi-modal model evaluation backend
    VLM_EVAL_KIT = 'VLMEvalKit'
    
    # Use RAGEval as the RAG evaluation backend
    RAG_EVAL = 'RAGEval'

    # Use third-party evaluation backend/modules
    THIRD_PARTY = 'ThirdParty'
    
    




import sys
from abc import abstractmethod
from typing import Any, Dict, Iterator, List, Tuple

import json

from evalscope.perf.arguments import Arguments


class DatasetPluginBase:

    def __init__(self, query_parameters: Arguments):
        """Build data set plugin

        Args:
            dataset_path (str, optional): The input dataset path. Defaults to None.
        """
        self.query_parameters = query_parameters

    def __next__(self):
        for item in self.build_messages():
            yield item
        raise StopIteration

    def __iter__(self):
        return self.build_messages()

    @abstractmethod
    def build_messages(self) -> Iterator[List[Dict]]:
        """Build the request.

        Raises:
            NotImplementedError: The request is not impletion.

        Yields:
            Iterator[List[Dict]]: Yield request messages.
        """
        raise NotImplementedError

    def dataset_line_by_line(self, dataset: str) -> Iterator[str]:
        """Get content line by line of dataset.

        Args:
            dataset (str): The dataset path.

        Yields:
            Iterator[str]: Each line of file.
        """
        with open(dataset, 'r', encoding='utf-8') as f:
            for line in f:
                yield line

    def dataset_json_list(self, dataset: str) -> Iterator[Dict]:
        """Read data from file which is list of requests.
           Sample: https://huggingface.co/datasets/Yukang/LongAlpaca-12k

        Args:
            dataset (str): The dataset path.

        Yields:
            Iterator[Dict]: The each request object.
        """
        with open(dataset, 'r', encoding='utf-8') as f:
            content = f.read()
        data = json.loads(content)
        for item in data:
            yield item



# Copyright (c) Alibaba, Inc. and its affiliates.
import os
from argparse import ArgumentParser

from evalscope.cli.base import CLICommand
from evalscope.perf.arguments import add_argument
from evalscope.perf.main import run_perf_benchmark

current_path = os.path.dirname(os.path.abspath(__file__))
root_path = os.path.dirname(current_path)


def subparser_func(args):
    """ Function which will be called for a specific sub parser.
    """
    return PerfBenchCMD(args)


class PerfBenchCMD(CLICommand):
    name = 'perf'

    def __init__(self, args):
        self.args = args

    @staticmethod
    def define_args(parsers: ArgumentParser):
        """ define args for create pipeline template command.
        """
        parser = parsers.add_parser(PerfBenchCMD.name)
        add_argument(parser)
        parser.set_defaults(func=subparser_func)

    def execute(self):
        run_perf_benchmark(self.args)



import base64
from io import BytesIO
from typing import Any, Dict, Iterator, List

from modelscope.msdatasets import MsDataset
from PIL import Image

from evalscope.perf.arguments import Arguments
from evalscope.perf.plugin.datasets.base import DatasetPluginBase
from evalscope.perf.plugin.registry import register_dataset


def PIL_to_base64(image: Image.Image) -> str:
    buffered = BytesIO()
    image.save(buffered, format='JPEG')
    img_str = base64.b64encode(buffered.getvalue()).decode('utf-8')
    return img_str


@register_dataset('flickr8k')
class FlickrDatasetPlugin(DatasetPluginBase):
    """Read dataset and return prompt.
    Datasets: https://www.modelscope.cn/datasets/clip-benchmark/wds_flickr8k/files
    """

    def __init__(self, query_parameters: Arguments):
        super().__init__(query_parameters)

    def build_messages(self) -> Iterator[List[Dict]]:
        dataset = MsDataset.load('clip-benchmark/wds_flickr8k', split='test')

        for item in dataset:
            pil_image = item['jpg']
            base64_iamge = PIL_to_base64(pil_image)

            yield [{
                'role':
                'user',
                'content': [
                    {
                        'type': 'text',
                        'text': 'Describe the image'
                    },
                    {
                        'type': 'image_url',
                        'image_url': {
                            'url': f'data:image/jpeg;base64,{base64_iamge}',
                        }
                    },
                ],
            }]



# Copyright (c) Alibaba, Inc. and its affiliates.
from dataclasses import dataclass

from swift.llm import (
    get_model_tokenizer, get_template, inference, get_default_template_type,
)
from swift.utils import seed_everything

# TODO: Support custom model for swift infer


@dataclass
class SwiftInferArgs:
    model_id_or_path: str
    model_type: str
    max_new_tokens: int = 2048


class SwiftInfer:

    def __init__(self, args: SwiftInferArgs):
        model_type = args.model_type
        template_type = get_default_template_type(model_type)
        model, tokenizer = get_model_tokenizer(model_type,
                                               model_id_or_path=args.model_id_or_path,
                                               model_kwargs={'device_map': 'auto'})
        model.generation_config.max_new_tokens = args.max_new_tokens
        print(f'** Generation config: {model.generation_config}')

        template = get_template(template_type, tokenizer)
        seed_everything(42)

        self.tokenizer = tokenizer
        self.model = model
        self.template = template

    def predict(self, system: str, query: str, history: list):

        response, history = inference(self.model,
                                      self.template,
                                      query=query,
                                      system=system,
                                      history=history)

        return response



import os
import asyncio
from datasets import Dataset
from evalscope.backend.rag_eval import EmbeddingModel, LLM
from evalscope.backend.rag_eval.ragas.tasks.translate_prompt import translate_prompts
from evalscope.utils.logger import get_logger
from .arguments import EvaluationArguments

logger = get_logger()


def rag_eval(
    args: EvaluationArguments,
) -> None:

    from ragas import evaluate, RunConfig
    from ragas.llms import LangchainLLMWrapper
    import importlib

    def dynamic_import(*function_names):
        functions = []
        for name in function_names:
            module = importlib.import_module('ragas.metrics')
            functions.append(getattr(module, name)())
        return functions

    llm = LLM.load(**args.critic_llm)
    embedding = EmbeddingModel.load(**args.embeddings)

    # load dataset
    dataset = Dataset.from_json(args.testset_file)

    # load metrics
    metrics = dynamic_import(*args.metrics)
    asyncio.run(
        translate_prompts(
            prompts=metrics,
            target_lang=args.language,
            llm=LangchainLLMWrapper(llm),
            adapt_instruction=True,
        )
    )

    # evaluate
    runconfig = RunConfig(timeout=600, max_retries=2, max_wait=60, max_workers=1)
    score = evaluate(
        dataset,
        metrics=metrics,
        llm=llm,
        embeddings=embedding,
        run_config=runconfig,
    )
    score_df = score.to_pandas()
    logger.info(score_df)

    output_path = args.testset_file.replace('.json', '_score.json')
    score_df.to_json(
        output_path, indent=4, index=False, orient='records', force_ascii=False
    )

    logger.info(f'Eval score saved to {output_path}')



import asyncio
import functools
import signal
import sys

from evalscope.utils.logger import get_logger

logger = get_logger()


def exception_handler(func):
    if asyncio.iscoroutinefunction(func):

        @functools.wraps(func)
        async def async_wrapper(*args, **kwargs):
            try:
                return await func(*args, **kwargs)
            except Exception as e:
                logger.exception(f"Exception in async function '{func.__name__}': {e}")
                sys.exit(1)

        return async_wrapper
    else:

        @functools.wraps(func)
        def sync_wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                logger.exception(f"Exception in function '{func.__name__}': {e}")
                sys.exit(1)

        return sync_wrapper


def signal_handler(signal_name, loop):
    logger.info('Got signal %s: exit' % signal_name)
    loop.stop()


def add_signal_handlers(loop):
    for signal_name in {'SIGINT', 'SIGTERM'}:
        loop.add_signal_handler(
            getattr(signal, signal_name),
            functools.partial(signal_handler, signal_name, loop),
        )



# Copyright (c) Alibaba, Inc. and its affiliates.

import re


def count_words(text):
    chinese_characters = re.findall(r'[\u4e00-\u9fff]', text)
    english_words = re.findall(r'\b[a-zA-Z]+\b', text)

    chinese_char_count = len(chinese_characters)
    english_word_count = len(english_words)

    total_count = chinese_char_count + english_word_count

    is_chinese = chinese_char_count > english_word_count

    return total_count, is_chinese


def chinese_to_arabic(chinese_number: str) -> int:
    chinese_numerals = {
        '零': 0,
        '一': 1,
        '二': 2,
        '三': 3,
        '四': 4,
        '五': 5,
        '六': 6,
        '七': 7,
        '八': 8,
        '九': 9,

        '俩': 2,
        '两': 2,
    }

    return chinese_numerals.get(chinese_number, chinese_number)



# Copyright (c) Alibaba, Inc. and its affiliates.

import os
import csv
from evalscope.benchmarks.data_adapter import DataAdapter
from evalscope.metrics.metrics import exact_match, weighted_mean
from evalscope.utils import ResponseParser, normalize_score
from evalscope.utils.logger import get_logger
# flake8: noqa

logger = get_logger()

DATASET_ID = 'modelscope/cmmlu'

SUBSET_LIST = [
    "agronomy",
    "anatomy",
    "ancient_chinese",
    "arts",
    "astronomy",
    "business_ethics",
    "chinese_civil_service_exam",
    "chinese_driving_rule",
    "chinese_food_culture",
    "chinese_foreign_policy",
    "chinese_history",
    "chinese_literature",
    "chinese_teacher_qualification",
    "college_actuarial_science",
    "college_education",
    "college_engineering_hydrology",
    "college_law",
    "college_mathematics",
    "college_medical_statistics",
    "clinical_knowledge",
    "college_medicine",
    "computer_science",
    "computer_security",
    "conceptual_physics",
    "construction_project_management",
    "economics",
    "education",
    "elementary_chinese",
    "elementary_commonsense",
    "elementary_information_and_technology",
    "electrical_engineering",
    "elementary_mathematics",
    "ethnology",
    "food_science",
    "genetics",
    "global_facts",
    "high_school_biology",
    "high_school_chemistry",
    "high_school_geography",
    "high_school_mathematics",
    "high_school_physics",
    "high_school_politics",
    "human_sexuality",
    "international_law",
    "journalism",
    "jurisprudence",
    "legal_and_moral_basis",
    "logical",
    "machine_learning",
    "management",
    "marketing",
    "marxist_theory",
    "modern_chinese",
    "nutrition",
    "philosophy",
    "professional_accounting",
    "professional_law",
    "professional_medicine",
    "professional_psychology",
    "public_relations",
    "security_study",
    "sociology",
    "sports_science",
    "traditional_chinese_medicine",
    "virology",
    "world_history",
    "world_religions"
]


SUBJECT_MAPPING = {"agronomy": ["other", "Other"],
                   "anatomy": ["biology", "STEM"],
                   "ancient_chinese": ["china specific", "China specific"],
                   "arts": ["arts", "Humanities"],
                   "astronomy": ["physics", "STEM"],
                   "business_ethics": ["business", "Social Science"],
                   "chinese_civil_service_exam": ["china specific", "China specific"],
                   "chinese_driving_rule": ["china specific", "China specific"],
                   "chinese_food_culture": ["china specific", "China specific"],
                   "chinese_foreign_policy": ["china specific", "China specific"],
                   "chinese_history": ["china specific", "China specific"],
                   "chinese_literature": ["china specific", "China specific"],
                   "chinese_teacher_qualification": ["china specific", "China specific"],
                   "college_actuarial_science": ["math", "STEM"],
                   "college_education": ["education", "Social Science"],
                   "college_engineering_hydrology": ["engineering", "STEM"],
                   "college_law": ["law", "Humanities"],
                   "college_mathematics": ["math", "STEM"],
                   "college_medical_statistics": ["statistics", "STEM"],
                   "clinical_knowledge": ["other", "Other"],
                   "college_medicine": ["other", "Other"],
                   "computer_science": ["computer science", "STEM"],
                   "computer_security": ["other", "Other"],
                   "conceptual_physics": ["physics", "STEM"],
                   "construction_project_management": ["china specific", "China specific"],
                   "economics": ["economics", "Social Science"],
                   "education": ["education", "Social Science"],
                   "elementary_chinese": ["china specific", "China specific"],
                   "elementary_commonsense": ["china specific", "China specific"],
                   "elementary_information_and_technology": ["other", "Other"],
                   "electrical_engineering": ["engineering", "STEM"],
                   "elementary_mathematics": ["math", "STEM"],
                   "ethnology": ["china specific", "China specific"],
                   "food_science": ["other", "Other"],
                   "genetics": ["biology", "STEM"],
                   "global_facts": ["global", "Humanities"],
                   "high_school_biology": ["biology", "STEM"],
                   "high_school_chemistry": ["chemistry", "STEM"],
                   "high_school_geography": ["geography", "Social Science"],
                   "high_school_mathematics": ["math", "STEM"],
                   "high_school_physics": ["physics", "STEM"],
                   "high_school_politics": ["china specific", "China specific"],
                   "human_sexuality": ["other", "Other"],
                   "international_law": ["law", "Humanities"],
                   "journalism": ["sociology", "Social Science"],
                   "jurisprudence": ["law", "Humanities"],
                   "legal_and_moral_basis": ["other", "Other"],
                   "logical": ["philosophy", "Humanities"],
                   "machine_learning": ["computer science", "STEM"],
                   "management": ["business", "Social Science"],
                   "marketing": ["business", "Social Science"],
                   "marxist_theory": ["philosophy", "Humanities"],
                   "modern_chinese": ["china specific", "China specific"],
                   "nutrition": ["other", "Other"],
                   "philosophy": ["philosophy", "Humanities"],
                   "professional_accounting": ["business", "Social Science"],
                   "professional_law": ["law", "Humanities"],
                   "professional_medicine": ["other", "Other"],
                   "professional_psychology": ["psychology", "Social Science"],
                   "public_relations": ["politics", "Social Science"],
                   "security_study": ["politics", "Social Science"],
                   "sociology": ["culture", "Social Science"],
                   "sports_science": ["other", "Other"],
                   "traditional_chinese_medicine": ["china specific", "China specific"],
                   "virology": ["biology", "STEM"],
                   "world_history": ["history", "Humanities"],
                   "world_religions": ["global", "Humanities"]
}


class CMMLUAdapter(DataAdapter):

    choices = ['A', 'B', 'C', 'D']

    def __init__(self,
                 subset_list: list = None,
                 metric_list: list = None,
                 few_shot_num: int = 5,
                 train_split: str = 'dev',
                 eval_split: str = 'test',
                 **kwargs):

        if subset_list is None:
            subset_list = SUBSET_LIST

        if metric_list is None:
            metric_list = [{'name': 'WeightedAverageAccuracy', 'object': weighted_mean}]

        super().__init__(subset_list=subset_list,
                        metric_list=metric_list,
                        few_shot_num=few_shot_num,
                        train_split=train_split,
                        eval_split=eval_split,
                        **kwargs)

    def load_from_disk(self, dataset_name_or_path, subset_list, work_dir, **kwargs) -> dict:
        data_dict = {}
        for subset_name in subset_list:
            data_dict[subset_name] = {}
            for split_name in [self.train_split, self.eval_split]:
                file_path = os.path.join(work_dir, dataset_name_or_path, split_name, f'{subset_name}.csv')
                if os.path.exists(file_path):
                    with open(file_path, encoding='utf-8') as f:
                        rows = []
                        reader = csv.reader(f)
                        for row in reader:
                            if len(row) != 7:
                                logger.error(f'Mismatch len of row: {row}, len of row should be 6. Skip this row.')
                                continue
                            rows.append({
                                'Question': row[1],
                                'A': row[2],
                                'B': row[3],
                                'C': row[4],
                                'D': row[5],
                                'Answer': row[6],
                            })

                        data_dict[subset_name].update({split_name: rows})

        return data_dict

    def gen_prompt(self, input_d: dict, subset_name: str, few_shot_list: list, **kwargs) -> dict:
        """
        Generate model prompt from raw input, unify the prompt format for CMMLU benchmark.

        Args:
            input_d (dict): The raw input. A single data format of the CMMLU:

            {'Question': '下列关于重力的说法正确的是',
            'A': '在地球周围的物体都要受到重力作用，与其运动状态无关',
            'B': '对某一物体而言，重力的大小是一个恒量，不随物体的地理位置而改变',
            'C': '重力就是地球对物体的吸引力，重力的方向总是竖直向下',
            'D': '在地球表面各处的重力方向都是相同的',
            'Answer': 'A'}

        Returns:
            {'data': [(context, continuation), ...]}

        """
        prompt = '以下是关于{}的单项选择题。\n\n'.format(
            self._format_subject(subset_name)
        )
        few_shot_prompts = [self._generate_prompt(input_d=sample, include_answer=True) for sample in few_shot_list]

        context: str = '\n'.join(few_shot_prompts) + '\n'
        context += self._generate_prompt(input_d=input_d, include_answer=False)
        context = prompt + context

        full_prompt: str = context.strip() + self._generate_prompt(input_d=input_d, include_answer=False)

        return {'data': [full_prompt], 'multi_choices': self.choices}

    def get_gold_answer(self, input_d: dict) -> str:
        # Get the gold choice
        return input_d.get('Answer', '')

    def parse_pred_result(self, result: str, raw_input_d: dict = None, eval_type: str = 'checkpoint') -> str:
        """
        Parse the model output to get the answer. Could be the best choice index.

        Args:
            result: Predicted answer from the model. Usually a string for chat.
            raw_input_d: The raw input. Depending on the dataset.
            eval_type: The evaluation type. 'checkpoint', 'service', 'custom'.

        Returns:
            The parsed answer. Depending on the dataset. Usually a string for chat.
        """
        if eval_type == 'checkpoint':
            return result
        elif eval_type == 'service':
            return ResponseParser.parse_first_option_with_choices(result, self.choices)  # TODO: to be checked !
        elif eval_type == 'custom':
            return ResponseParser.parse_first_option_with_choices(result, self.choices)  # TODO: to be checked !
        else:
            raise ValueError(f'Invalid eval_type: {eval_type}')

    def match(self, gold: str, pred: str) -> float:
        return exact_match(gold=gold, pred=pred)

    def compute_metric(self, review_res_list: list) -> float:
        """
        Compute evaluation result by specific metric.

        Args:
            review_res_list: review score list, e.g. [0, 1, 1, 0, ...]

        Returns:
            The metric score.
        """
        items = [(score, 1.0) for score in review_res_list]
        return weighted_mean(items)

    def gen_report(self, subset_score_map: dict, report_name: str = None) -> dict:
        """
        Generate report for the evaluation.

        Args:
            subset_score_map: The subset-score mapping. e.g. {subset_name: (score, num), ...}
            report_name: the user-defined report name. Default: None

        Returns:
        {
            "name":"CMMLU",
            "metric":"WeightedAverageAccuracy",
            "score":0.3389,
            "category":[
                {
                   "name":"STEM",
                   "score":0.2528,
                   "subset":[
                       {
                           "name":"computer_network",
                           "score":0.2632
                       },
                       {
                           "name":"operating_system",
                           "score":0.3157
                       },
                       {
                           "name":"computer_architecture",
                           "score":0.4285
                       }
                   ]
                }
            ],
            "total_num":59
        }
        """
        total_num: int = sum([num for _, num in subset_score_map.values()])
        weighted_avg_acc: float = sum([score * num for score, num in subset_score_map.values()]) / total_num

        # Get domain-subject mapping
        subject_review_map = {}
        for subset_name, (subset_score, num) in subset_score_map.items():
            domain_name: str = SUBJECT_MAPPING.get(subset_name)[1] if SUBJECT_MAPPING.get(subset_name) else subset_name
            if domain_name in subject_review_map:
                subject_review_map[domain_name].append((subset_name, subset_score, num))
            else:
                subject_review_map[domain_name] = [(subset_name, subset_score, num)]

        # Get domain score
        category_list = []
        for domain_name, domain_res_list in subject_review_map.items():
            domain_weighted_avg_acc = sum([score * num for _, score, num in domain_res_list]) / \
                                     sum([num for _, _, num in domain_res_list])
            domain_weighted_avg_acc = normalize_score(score=domain_weighted_avg_acc)
            category_list.append({'name': domain_name,
                                  'score': domain_weighted_avg_acc,
                                  'subset': [{'name': subset_name, 'score': normalize_score(subset_score)}
                                             for subset_name, subset_score, _ in domain_res_list]})

        # Get final dict of report
        res_map = dict(name=report_name or 'cmmlu',
                       metric=self.metric_list[0]['name'],
                       score=weighted_avg_acc,
                       category=category_list,
                       total_num=total_num)

        return res_map

    @classmethod
    def _generate_prompt(cls, input_d: dict, include_answer=True) -> str:

        input_choices: list = [input_d['A'], input_d['B'], input_d['C'], input_d['D']]

        example: str = input_d['Question']
        for j in range(len(cls.choices)):
            example += '\n{}. {}'.format(cls.choices[j], input_choices[j])

        example += '\nAnswer:'
        if include_answer:
            example += ' {}\n\n'.format(input_d['Answer'])

        return example

    @classmethod
    def _format_subject(cls, subject):
        l = subject.split('_')
        s = ''
        for entry in l:
            s += ' ' + entry
        return s


# Copyright (c) Alibaba, Inc. and its affiliates.

from evalscope.benchmarks.trivia_qa.trivia_qa_adapter import TriviaQaAdapter, DATASET_ID, SUBSET_LIST
from evalscope.benchmarks.trivia_qa.trivia_qa_adapter import TriviaQaAdapter as DataAdapterClass
from evalscope.models.model_adapter import ChatGenerationModelAdapter as ModelAdapterClass    # noqa



import itertools

from datasets import Dataset, DatasetDict

from mteb.abstasks.AbsTaskClustering import AbsTaskClustering
from mteb.abstasks.AbsTaskClusteringFast import (
    AbsTaskClusteringFast,
    check_label_distribution,
)
from mteb.abstasks.TaskMetadata import TaskMetadata

NUM_SAMPLES = 2048


class CLSClusteringFastS2S(AbsTaskClusteringFast):
    max_document_to_embed = NUM_SAMPLES
    max_fraction_of_documents_to_embed = None

    metadata = TaskMetadata(
        name='CLSClusteringS2S',
        description='Clustering of titles from CLS dataset. Clustering of 13 sets on the main category.',
        reference='https://arxiv.org/abs/2209.05034',
        dataset={
            'path': 'C-MTEB/CLSClusteringS2S',
            'revision': 'e458b3f5414b62b7f9f83499ac1f5497ae2e869f',
        },
        type='Clustering',
        category='s2s',
        modalities=['text'],
        eval_splits=['test'],
        eval_langs=['cmn-Hans'],
        main_score='v_measure',
        date=('2022-01-01', '2022-09-12'),
        domains=['Academic', 'Written'],
        task_subtypes=['Thematic clustering', 'Topic classification'],
        license='apache-2.0',
        annotations_creators='derived',
        dialect=[],
        sample_creation='found',
        bibtex_citation="""@misc{li2022csl,
            title={CSL: A Large-scale Chinese Scientific Literature Dataset},
            author={Yudong Li and Yuqing Zhang and Zhe Zhao and Linlin Shen and Weijie Liu and Weiquan Mao and Hui Zhang},
            year={2022},
            eprint={2209.05034},
            archivePrefix={arXiv},
            primaryClass={cs.CL}
        }""",  # noqa
        descriptive_stats={
            'n_samples': {'test': NUM_SAMPLES},
            'avg_character_length': {},
        },
    )

    def dataset_transform(self):
        ds = {}
        for split in self.metadata.eval_splits:
            labels = list(itertools.chain.from_iterable(self.dataset[split]['labels']))
            sentences = list(
                itertools.chain.from_iterable(self.dataset[split]['sentences'])
            )

            check_label_distribution(self.dataset[split])

            ds[split] = Dataset.from_dict({'labels': labels, 'sentences': sentences})
        self.dataset = DatasetDict(ds)
        self.dataset = self.stratified_subsampling(
            self.dataset,
            self.seed,
            self.metadata.eval_splits,
            label='labels',
            n_samples=NUM_SAMPLES,
        )


class CLSClusteringFastP2P(AbsTaskClusteringFast):
    max_document_to_embed = NUM_SAMPLES
    max_fraction_of_documents_to_embed = None

    metadata = TaskMetadata(
        name='CLSClusteringP2P',
        description='Clustering of titles + abstract from CLS dataset. Clustering of 13 sets on the main category.',
        reference='https://arxiv.org/abs/2209.05034',
        dataset={
            'path': 'C-MTEB/CLSClusteringP2P',
            'revision': '4b6227591c6c1a73bc76b1055f3b7f3588e72476',
        },
        type='Clustering',
        category='p2p',
        modalities=['text'],
        eval_splits=['test'],
        eval_langs=['cmn-Hans'],
        main_score='v_measure',
        date=('2022-01-01', '2022-09-12'),
        domains=['Academic', 'Written'],
        task_subtypes=['Thematic clustering', 'Topic classification'],
        license='apache-2.0',
        annotations_creators='derived',
        dialect=[],
        sample_creation='found',
        bibtex_citation="""@misc{li2022csl,
            title={CSL: A Large-scale Chinese Scientific Literature Dataset},
            author={Yudong Li and Yuqing Zhang and Zhe Zhao and Linlin Shen and Weijie Liu and Weiquan Mao and Hui Zhang},
            year={2022},
            eprint={2209.05034},
            archivePrefix={arXiv},
            primaryClass={cs.CL}
        }""",  # noqa
        descriptive_stats={
            'n_samples': {'test': NUM_SAMPLES},
            'avg_character_length': {},
        },
    )

    def dataset_transform(self):
        ds = {}
        for split in self.metadata.eval_splits:
            labels = list(itertools.chain.from_iterable(self.dataset[split]['labels']))
            sentences = list(
                itertools.chain.from_iterable(self.dataset[split]['sentences'])
            )

            check_label_distribution(self.dataset[split])

            ds[split] = Dataset.from_dict({'labels': labels, 'sentences': sentences})
        self.dataset = DatasetDict(ds)
        self.dataset = self.stratified_subsampling(
            self.dataset,
            self.seed,
            self.metadata.eval_splits,
            label='labels',
            n_samples=NUM_SAMPLES,
        )


class ThuNewsClusteringFastS2S(AbsTaskClusteringFast):
    max_document_to_embed = NUM_SAMPLES
    max_fraction_of_documents_to_embed = None

    metadata = TaskMetadata(
        name='ThuNewsClusteringS2S',
        dataset={
            'path': 'C-MTEB/ThuNewsClusteringS2S',
            'revision': '8a8b2caeda43f39e13c4bc5bea0f8a667896e10d',
        },
        description='Clustering of titles from the THUCNews dataset',
        reference='http://thuctc.thunlp.org/',
        type='Clustering',
        category='s2s',
        modalities=['text'],
        eval_splits=['test'],
        eval_langs=['cmn-Hans'],
        main_score='v_measure',
        date=('2006-01-01', '2007-01-01'),
        domains=['News', 'Written'],
        task_subtypes=['Thematic clustering', 'Topic classification'],
        license='apache-2.0',
        annotations_creators='derived',
        dialect=[],
        sample_creation='found',
        bibtex_citation="""@software{THUCTC,
  author = {Sun, M. and Li, J. and Guo, Z. and Yu, Z. and Zheng, Y. and Si, X. and Liu, Z.},
  title = {THUCTC: An Efficient Chinese Text Classifier},
  year = {2016},
  note = {THU Chinese Text Classification Toolkit},
  publisher = {THU Natural Language Processing Lab},
  url = {https://github.com/thunlp/THUCTC}
}""",
        descriptive_stats={
            'n_samples': {'test': NUM_SAMPLES},
            'avg_character_length': {},
        },
    )

    def dataset_transform(self):
        ds = {}
        for split in self.metadata.eval_splits:
            labels = list(itertools.chain.from_iterable(self.dataset[split]['labels']))
            sentences = list(
                itertools.chain.from_iterable(self.dataset[split]['sentences'])
            )

            check_label_distribution(self.dataset[split])

            ds[split] = Dataset.from_dict({'labels': labels, 'sentences': sentences})
        self.dataset = DatasetDict(ds)
        self.dataset = self.stratified_subsampling(
            self.dataset,
            self.seed,
            self.metadata.eval_splits,
            label='labels',
            n_samples=NUM_SAMPLES,
        )


class ThuNewsClusteringFastP2P(AbsTaskClusteringFast):
    max_document_to_embed = NUM_SAMPLES
    max_fraction_of_documents_to_embed = None

    metadata = TaskMetadata(
        name='ThuNewsClusteringP2P',
        dataset={
            'path': 'C-MTEB/ThuNewsClusteringP2P',
            'revision': '5798586b105c0434e4f0fe5e767abe619442cf93',
        },
        description='Clustering of titles + abstracts from the THUCNews dataset',
        reference='http://thuctc.thunlp.org/',
        type='Clustering',
        category='p2p',
        modalities=['text'],
        eval_splits=['test'],
        eval_langs=['cmn-Hans'],
        main_score='v_measure',
        date=('2006-01-01', '2007-01-01'),
        domains=['News', 'Written'],
        task_subtypes=['Thematic clustering', 'Topic classification'],
        license='apache-2.0',
        annotations_creators='derived',
        dialect=[],
        sample_creation='found',
        bibtex_citation="""@software{THUCTC,
  author = {Sun, M. and Li, J. and Guo, Z. and Yu, Z. and Zheng, Y. and Si, X. and Liu, Z.},
  title = {THUCTC: An Efficient Chinese Text Classifier},
  year = {2016},
  note = {THU Chinese Text Classification Toolkit},
  publisher = {THU Natural Language Processing Lab},
  url = {https://github.com/thunlp/THUCTC}
}""",
        descriptive_stats={
            'n_samples': {'test': NUM_SAMPLES},
            'avg_character_length': {},
        },
    )

    def dataset_transform(self):
        ds = {}
        for split in self.metadata.eval_splits:
            labels = list(itertools.chain.from_iterable(self.dataset[split]['labels']))
            sentences = list(
                itertools.chain.from_iterable(self.dataset[split]['sentences'])
            )

            check_label_distribution(self.dataset[split])

            ds[split] = Dataset.from_dict({'labels': labels, 'sentences': sentences})
        self.dataset = DatasetDict(ds)
        self.dataset = self.stratified_subsampling(
            self.dataset,
            self.seed,
            self.metadata.eval_splits,
            label='labels',
            n_samples=NUM_SAMPLES,
        )
